#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for prescription based pipeline units."""

import abc
import re
import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import Union
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from voluptuous import Schema
from voluptuous import Required

import attr

from thoth.adviser.exceptions import NotAcceptable
from thoth.adviser.exceptions import EagerStopPipeline
from thoth.adviser.state import State
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion

from .unit_cache import should_include_cache
from ...unit import Unit

if TYPE_CHECKING:
    from typing import Callable
    from ...pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


class _ValueList:
    """A class that overrides `in` to transparently handle included and excluded values."""

    __slots__ = ["_list"]

    def __init__(self, obj: Union[List[object], Dict[str, List[object]]]) -> None:
        """Initialize self."""
        self._list = obj

    def __contains__(self, item: str) -> bool:
        """Override default in behavior based on the YAML definition."""
        if isinstance(self._list, list):
            return self._list.__contains__(item)

        return not self._list["not"].__contains__(item)


@attr.s(slots=True)
class UnitPrescription(Unit, metaclass=abc.ABCMeta):
    """A base class for implementing pipeline units based on prescription supplied."""

    # Each prescription unit defines these specifically.
    SHOULD_INCLUDE_CACHE: Dict[str, bool] = {}
    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): str,
            Required("match"): object,
            Required("run"): object,
        }
    )

    _PRESCRIPTION: Optional[Dict[str, Any]] = None

    _stack_info_run = attr.ib(type=bool, kw_only=True, default=False)
    _configuration = attr.ib(type=Dict[str, Any], kw_only=True)
    prescription = attr.ib(type=Dict[str, Any], kw_only=True)

    @prescription.default
    def _prescription_default(self) -> Dict[str, Any]:
        """Initialize prescription property."""
        if self._PRESCRIPTION is None:
            raise ValueError("No assigned prescription on the class level to be set")

        return self._PRESCRIPTION

    @property
    def run_prescription(self) -> Dict[str, Any]:
        """Get run part of the prescription assigned."""
        return self._configuration.get("run", {})

    @property
    def match_prescription(self) -> Dict[str, Any]:
        """Get match part of the prescription assigned."""
        return self._configuration.get("match", {})

    @_configuration.default
    def _initialize_default_configuration(self) -> Dict[str, Any]:
        """Initialize default unit configuration based on declared class' default configuration."""
        if self._PRESCRIPTION is None:
            raise ValueError("No assigned prescription on the class level to be set")

        return {
            "package_name": None,
            "match": self._PRESCRIPTION.get("match", {}),
            "run": self._PRESCRIPTION.get("run", {}),
        }

    @classmethod
    def get_unit_name(cls) -> str:
        """Get the name of the current prescription unit.

        This method is a class method and *MUST NOT* be used when obtaining unit name on an
        instance. As part of the memory optimization we use class to get the current name of
        a prescription unit with assigned prescription. This means that the prescription unit
        instance would have different names reported with this method based on the current
        class context.
        """
        if cls._PRESCRIPTION is None:
            raise ValueError("No prescription defined")

        name: str = cls._PRESCRIPTION["name"]
        return name

    @classmethod
    def set_prescription(cls, prescription: Dict[str, Any]) -> None:
        """Set prescription to the unit."""
        cls._PRESCRIPTION = prescription

    @classmethod
    def _check_symbols(
        cls, unit_name: str, library_name: str, symbols_expected: List[str], symbols_used: List[str]
    ) -> bool:
        """Check if symbols expected are available given the symbols used."""
        for symbol_expected in symbols_expected:
            for symbol_used in symbols_used:
                if symbol_expected.endswith(".*"):
                    if symbol_used.startswith(symbol_expected[:-2]):  # Discard ending ".*"
                        _LOGGER.debug(
                            "%s: Library symbol %r matching unit requirement on symbol %r for %r",
                            unit_name,
                            symbol_used,
                            symbol_expected,
                            library_name,
                        )
                        break
                elif symbol_used == symbol_expected:
                    _LOGGER.debug(
                        "%s: Library symbol %r matching unit requirement on symbol %r for %r",
                        unit_name,
                        symbol_used,
                        symbol_expected,
                        library_name,
                    )
                    break

                _LOGGER.debug(
                    "%s: Not registering as library symbol requested %r for %r is not used",
                    unit_name,
                    symbol_expected,
                    library_name,
                )
                return False

        _LOGGER.debug("%s: All library symbols required for %r unit are used", unit_name, library_name)
        return True

    @staticmethod
    def _check_version(version_present: Optional[str], version_spec_declared: Optional[str]) -> bool:
        """Check that version present matches version specification."""
        if version_present is None:
            if version_spec_declared is not None:
                return False
            else:
                return True
        else:
            if version_spec_declared is None:
                return True

            return Version(version_present) in SpecifierSet(version_spec_declared)

    @classmethod
    def _should_include_base(cls, builder_context: "PipelineBuilderContext") -> bool:
        """Determine if this unit should be included."""
        if cls._PRESCRIPTION is None:
            raise ValueError("No prescription defined")

        should_include_dict = cls._PRESCRIPTION["should_include"]
        unit_name = cls.get_unit_name()

        times = should_include_dict.get("times", 1)  # XXX: We allow values 0 or 1 in the schema described.
        if times == 0 or builder_context.is_included(cls):
            return False

        if not cls._should_include_base_cached(unit_name, builder_context, should_include_dict):
            # Using pre-cached results based on parts that do not change or first time run.
            return False

        # Dependencies.
        dependencies = should_include_dict.get("dependencies", {})
        for boot_name in dependencies.get("boots", []):
            if boot_name not in builder_context.get_included_boot_names():
                _LOGGER.debug("%s: Not registering as dependency on boot %r is not satisfied", unit_name, boot_name)
                return False

        for pseudonym_name in dependencies.get("pseudonyms", []):
            if pseudonym_name not in builder_context.get_included_pseudonym_names():
                _LOGGER.debug(
                    "%s: Not registering as dependency on pseudonym %r is not satisfied", unit_name, pseudonym_name
                )
                return False

        for sieve_name in dependencies.get("sieves", []):
            if sieve_name not in builder_context.get_included_sieve_names():
                _LOGGER.debug("%s: Not registering as dependency on sieve %r is not satisfied", unit_name, sieve_name)
                return False

        for step_name in dependencies.get("steps", []):
            if step_name not in builder_context.get_included_step_names():
                _LOGGER.debug("%s: Not registering as dependency on step %r is not satisfied", unit_name, step_name)
                return False

        for stride_name in dependencies.get("strides", []):
            if stride_name not in builder_context.get_included_stride_names():
                _LOGGER.debug("%s: Not registering as dependency on stride %r is not satisfied", unit_name, stride_name)
                return False

        for wrap_name in dependencies.get("wraps", []):
            if wrap_name not in builder_context.get_included_wrap_names():
                _LOGGER.debug("%s: Not registering as dependency on stride %r is not satisfied", unit_name, wrap_name)
                return False

        return True

    if TYPE_CHECKING:
        SHOULD_INCLUDE_FUNC_TYPE = Callable[
            [Type["UnitPrescription"], str, "PipelineBuilderContext", Dict[str, Any]], bool
        ]

    @classmethod
    @should_include_cache
    def _should_include_base_cached(
        cls, unit_name: str, builder_context: "PipelineBuilderContext", should_include_dict: Dict[str, Any]
    ) -> bool:
        """Determine if this unit should be included."""
        adviser_pipeline = should_include_dict.get("adviser_pipeline", False)
        if not adviser_pipeline and builder_context.is_adviser_pipeline():
            _LOGGER.debug("%s: Not registering for adviser pipeline", unit_name)
            return False
        elif adviser_pipeline and builder_context.is_adviser_pipeline():
            allowed_recommendation_types = should_include_dict.get("recommendation_types")
            if (
                allowed_recommendation_types is not None
                and builder_context.recommendation_type is not None
                and builder_context.recommendation_type.name.lower() not in _ValueList(allowed_recommendation_types)
            ):
                _LOGGER.debug(
                    "%s: Not registering for adviser pipeline with recommendation type %s",
                    unit_name,
                    builder_context.recommendation_type.name,
                )
                return False

        if (
            not should_include_dict.get("dependency_monkey_pipeline", False)
            and builder_context.is_dependency_monkey_pipeline()
        ):
            _LOGGER.debug("%s: Not registering for dependency monkey pipeline", unit_name)
            return False
        elif (
            should_include_dict.get("dependency_monkey_pipeline", False)
            and builder_context.is_dependency_monkey_pipeline()
        ):
            allowed_decision_types = should_include_dict.get("decision_types")
            if (
                allowed_decision_types is not None
                and builder_context.decision_type is not None
                and builder_context.decision_type.name.lower() not in _ValueList(allowed_decision_types)
            ):
                _LOGGER.debug(
                    "%s: Not registering for dependency monkey pipeline with decision type %s",
                    unit_name,
                    builder_context.decision_type.name,
                )
                return False

        labels_expected = should_include_dict.get("labels", {})
        if labels_expected:
            for label_key, value in labels_expected.items():
                value_context = builder_context.labels.get(label_key)
                if value == value_context:
                    break
            else:
                _LOGGER.debug(
                    "%s: Not registering as labels requested %s do not match with labels supplied %s",
                    unit_name,
                    labels_expected,
                    builder_context.labels,
                )
                return False

        # Library usage.
        library_usage_expected = should_include_dict.get("library_usage", {})
        if library_usage_expected:
            if not builder_context.library_usage:
                _LOGGER.debug("%s: Not registering as no library usage supplied", unit_name)
                return False

            for library_name, symbols_expected in library_usage_expected.items():
                symbols_used = builder_context.library_usage.get(library_name)

                if not symbols_used:
                    _LOGGER.debug("%s: Not registering as library %s is not used", unit_name, library_name)
                    return False

                if not cls._check_symbols(unit_name, library_name, symbols_expected, symbols_used):
                    return False
            else:
                _LOGGER.debug("%s: All library symbols required present in the library usage supplied", unit_name)

        runtime_environment_dict = should_include_dict.get("runtime_environments", {})

        # Operating system.
        operating_systems = runtime_environment_dict.get("operating_systems")
        os_used = builder_context.project.runtime_environment.operating_system
        os_used_name = os_used.name if os_used is not None else None
        os_used_version = os_used.version if os_used is not None else None

        if operating_systems:
            for item in operating_systems:
                os_name = item.get("name")
                os_version = item.get("version")
                if os_name == os_used_name and os_version == os_used_version:
                    _LOGGER.debug("%s: Matching operating system %r in version %r", unit_name, os_name, os_version)
                    break
            else:
                _LOGGER.debug(
                    "%s: Not matching operating system (using %r in version %r)",
                    unit_name,
                    os_used_name,
                    os_used_version,
                )
                return False

        # Hardware.
        hw_used = builder_context.project.runtime_environment.hardware

        for hardware_dict in runtime_environment_dict.get("hardware", []):
            # CPU/GPU
            cpu_families = hardware_dict.get("cpu_families")
            cpu_models = hardware_dict.get("cpu_models")
            gpu_models = hardware_dict.get("gpu_models")
            if cpu_families is not None and hw_used.cpu_family not in _ValueList(cpu_families):
                _LOGGER.debug("%s: Not matching CPU family used (using %r)", unit_name, hw_used.cpu_family)
                return False

            if cpu_models is not None and hw_used.cpu_model not in _ValueList(cpu_models):
                _LOGGER.debug("%s: Not matching CPU model used (using %r)", unit_name, hw_used.cpu_model)
                return False

            if gpu_models is not None and hw_used.gpu_model not in _ValueList(gpu_models):
                _LOGGER.debug("%s: Not matching GPU model used (using %r)", unit_name, hw_used.gpu_model)
                return False

        # Software present.
        runtime_used = builder_context.project.runtime_environment

        python_version_spec = runtime_environment_dict.get("python_version")
        if not cls._check_version(runtime_used.python_version, python_version_spec):
            _LOGGER.debug(
                "%s: Not matching Python version used (using %r; expected %r)",
                unit_name,
                runtime_used.python_version,
                python_version_spec,
            )
            return False

        cuda_version_spec = runtime_environment_dict.get("cuda_version")
        if not cls._check_version(runtime_used.cuda_version, cuda_version_spec):
            _LOGGER.debug(
                "%s: Not matching CUDA version used (using %r; expected %r)",
                unit_name,
                runtime_used.cuda_version,
                cuda_version_spec,
            )
            return False

        platforms = runtime_environment_dict.get("platforms")
        if platforms is not None and runtime_used.platform not in _ValueList(platforms):
            _LOGGER.debug("%s: Not matching platform used (using %r)", unit_name, runtime_used.platform)
            return False

        openblas_version_spec = runtime_environment_dict.get("openblas_version")
        if not cls._check_version(runtime_used.openblas_version, openblas_version_spec):
            _LOGGER.debug(
                "%s: Not matching openblas version used (using %r; expected %r)",
                unit_name,
                runtime_used.openblas_version,
                openblas_version_spec,
            )
            return False

        openmpi_version_spec = runtime_environment_dict.get("openmpi_version")
        if not cls._check_version(runtime_used.openmpi_version, openmpi_version_spec):
            _LOGGER.debug(
                "%s: Not matching openmpi version used (using %r; expected %r)",
                unit_name,
                runtime_used.openmpi_version,
                openmpi_version_spec,
            )
            return False

        cudnn_version_spec = runtime_environment_dict.get("cudnn_version")
        if not cls._check_version(runtime_used.cudnn_version, cudnn_version_spec):
            _LOGGER.debug(
                "%s: Not matching cudnn version used (using %r; expected %r)",
                unit_name,
                runtime_used.cudnn_version,
                cudnn_version_spec,
            )
            return False

        mkl_version_spec = runtime_environment_dict.get("mkl_version")
        if not cls._check_version(runtime_used.mkl_version, mkl_version_spec):
            _LOGGER.debug(
                "%s: Not matching mkl version used (using %r; expected %r)",
                unit_name,
                runtime_used.mkl_version,
                mkl_version_spec,
            )
            return False

        base_images = runtime_environment_dict.get("base_images")
        if base_images is not None and runtime_used.base_image not in _ValueList(base_images):
            _LOGGER.debug("%s: Not matching base image used (using %r)", unit_name, runtime_used.base_image)
            return False

        # All the following require base image information.
        base_image = None
        if runtime_used.base_image:
            base_image = cls.get_base_image(runtime_used.base_image, raise_on_error=True)

        abi = runtime_environment_dict.get("abi")
        if abi:
            if not base_image:
                _LOGGER.debug(
                    "%s: Check on ABI present but no base image provided",
                    unit_name,
                )
                return False

            symbols_present = set(
                builder_context.graph.get_thoth_s2i_analyzed_image_symbols_all(
                    base_image[0],
                    base_image[1],
                    is_external=False,
                )
            )
            if not symbols_present:
                if builder_context.iteration == 0:
                    _LOGGER.warning(
                        f"%s: No symbols found for runtime environment %r", unit_name, runtime_used.base_image
                    )
                return False

            if isinstance(abi, dict) and "not" in abi:
                # Negate operation.
                if symbols_present.intersection(set(abi["not"])):
                    _LOGGER.debug("%s: Not matching ABI present in the base image", unit_name)
                    return False
                else:
                    return True
            elif isinstance(abi, list):
                if set(abi).issubset(symbols_present):
                    return True
                else:
                    _LOGGER.debug("%s: Not matching ABI present in the base image", unit_name)
                    return False
            else:
                _LOGGER.error("%s: Unknown ABI definition - please report this error to administrator", unit_name)
                return False

        rpm_packages = runtime_environment_dict.get("rpm_packages")
        if rpm_packages:
            if not base_image:
                _LOGGER.debug(
                    "%s: Check on RPM packages present but no base image provided",
                    unit_name,
                )
                return False

            analysis_document_id = builder_context.graph.get_last_analysis_document_id(
                base_image[0],
                base_image[1],
                is_external=False,
            )

            if not analysis_document_id:
                if builder_context.iteration == 0:
                    _LOGGER.warning(
                        "%s: No analysis for base container image %r found",
                        unit_name,
                        runtime_used.base_image,
                    )
                return False

            rpm_packages_present = builder_context.graph.get_rpm_package_version_all(analysis_document_id)
            if not rpm_packages_present:
                _LOGGER.debug("%s: No RPM packages found for %r", unit_name, runtime_used.base_image)
                return False
            if not cls._check_rpm_packages(unit_name, rpm_packages_present, rpm_packages):
                _LOGGER.debug(
                    "%s: Not matching RPM packages present in the base image %r", unit_name, runtime_used.base_image
                )
                return False

        python_packages = runtime_environment_dict.get("python_packages")
        if python_packages:
            if not base_image:
                _LOGGER.debug(
                    "%s: Check on Python packages present but no base image provided",
                    unit_name,
                )
                return False

            analysis_document_id = builder_context.graph.get_last_analysis_document_id(
                base_image[0],
                base_image[1],
                is_external=False,
            )

            if not analysis_document_id:
                if builder_context.iteration == 0:
                    _LOGGER.warning(
                        "%s: No analysis for base container image %r found",
                        unit_name,
                        runtime_used.base_image,
                    )
                return False

            python_packages_present = builder_context.graph.get_python_package_version_all(analysis_document_id)
            if not python_packages_present:
                _LOGGER.debug("%s: No Python packages found for %r", unit_name, runtime_used.base_image)
                return False
            if not cls._check_python_packages(unit_name, python_packages_present, python_packages):
                _LOGGER.debug(
                    "%s: Not matching Python packages present in the base image %r", unit_name, runtime_used.base_image
                )
                return False

        return True

    @classmethod
    def _index_url_check(cls, index_url_conf: Optional[Union[str, Dict[str, str]]], index_url: str) -> bool:
        """Convert index_url to a comparable object considering "not"."""
        if index_url_conf is None:
            return True

        if isinstance(index_url_conf, dict):
            if list(index_url_conf.keys()) != ["not"]:
                raise ValueError("index_url configuration should state directly string or a 'not' value")

            return index_url_conf["not"] != index_url
        else:
            return index_url_conf == index_url

    @classmethod
    def _check_python_packages(
        cls,
        unit_name: str,
        python_packages_present: List[Dict[str, str]],
        python_packages_required: List[Dict[str, str]],
    ) -> bool:
        """Check if required Python packages are present in the environment."""
        # Convert to dict to have O(1) access time.
        py_packages_present_dict: Dict[str, List[Dict[str, str]]] = {}
        for python_package_present in python_packages_present:
            package = py_packages_present_dict.get(python_package_present["package_name"])
            if package is None:
                py_packages_present_dict[python_package_present["package_name"]] = [python_package_present]
            else:
                package.append(python_package_present)

        if isinstance(python_packages_required, dict):
            if "not" not in python_packages_required:
                _LOGGER.error("%s: Unable to parse description of Python packages required", unit_name)
                return False

            for not_required_python_package in python_packages_required["not"]:
                for py_package_present in py_packages_present_dict.get(not_required_python_package["name"]) or []:
                    location = not_required_python_package.get("location")
                    if location is not None and not re.fullmatch(location, py_package_present["location"]):
                        _LOGGER.debug(
                            "%s: Python package %r in %r is located in different location %r as expected",
                            unit_name,
                            not_required_python_package["name"],
                            py_package_present["location"],
                            location,
                        )
                        continue

                    version = not_required_python_package.get("version")
                    if version and py_package_present["package_version"] not in SpecifierSet(version):
                        _LOGGER.debug(
                            "%s: Python package '%s==%s' (in %r) matches version %r but should not",
                            unit_name,
                            not_required_python_package["name"],
                            py_package_present["package_version"],
                            py_package_present["location"],
                            version,
                        )
                        continue

                    _LOGGER.debug(
                        "%s: presence of Python package %r causes not including the pipeline unit",
                        unit_name,
                        py_package_present,
                    )
                    return False
        else:
            for required_python_package in python_packages_required:
                for py_package_present in py_packages_present_dict.get(required_python_package["name"]) or []:
                    version = required_python_package.get("version")
                    if version and py_package_present["package_version"] not in SpecifierSet(version):
                        _LOGGER.debug(
                            "%s: Python package '%s==%s' (in %r) does not match required version %r",
                            unit_name,
                            required_python_package["name"],
                            py_package_present["package_version"],
                            py_package_present.get("location", "any"),
                            version,
                        )
                        continue

                    location = required_python_package.get("location")
                    if location is not None and not re.fullmatch(location, py_package_present["location"]):
                        _LOGGER.debug(
                            "%s: Python package %r is located at %r but expected to be in %r",
                            unit_name,
                            required_python_package["name"],
                            py_package_present["location"],
                            location,
                        )
                        continue

                    _LOGGER.debug(
                        "%s: Python package %r in version %r (located in %r) is found in the runtime environment",
                        unit_name,
                        required_python_package["name"],
                        required_python_package.get("version", "any"),
                        py_package_present.get("location", "any"),
                    )
                    break
                else:
                    _LOGGER.debug(
                        "%s: Not including as Python package %r (in version %r) is not present in the environment",
                        unit_name,
                        required_python_package["name"],
                        required_python_package.get("version", "any"),
                    )
                    return False

        _LOGGER.debug("%s: all Python package presence checks passed", unit_name)
        return True

    @staticmethod
    def _check_rpm_packages(
        unit_name: str,
        rpm_packages_present: List[Dict[str, str]],
        rpm_packages_required: Union[List[Dict[str, str]], Dict[str, List[Dict[str, str]]]],
    ) -> bool:
        """Check if required RPM packages are present."""
        # Convert RPM packages present to mapping to save some cycles and have O(1) look up.
        rpm_packages_pres = {i["package_name"]: i for i in rpm_packages_present}
        rpm_packages_req: List[Dict[str, str]]
        if isinstance(rpm_packages_required, dict):
            if "not" not in rpm_packages_required:
                _LOGGER.error("%s: Unable to parse description of RPM packages required", unit_name)
                return False

            should_be_present = False
            rpm_packages_req = [i for i in rpm_packages_required["not"]]
        else:
            should_be_present = True
            rpm_packages_req = [i for i in rpm_packages_required]

        for rpm_package_req in rpm_packages_req:
            rpm_name = rpm_package_req["package_name"]
            rpm_present = rpm_packages_pres.get(rpm_name)

            if should_be_present:
                if not rpm_present:
                    _LOGGER.debug(
                        "%s: Not including unit as RPM %r is not present in the runtime environment",
                        unit_name,
                        rpm_name,
                    )
                    return False

                for key, value in rpm_package_req.items():
                    value_present = rpm_present.get(key)
                    if value_present != value:
                        _LOGGER.debug(
                            "%s: Not including unit as RPM %r has not matching %r - expected %r got %r",
                            unit_name,
                            rpm_name,
                            key,
                            value,
                            value_present,
                        )
                        return False
            else:
                if not rpm_present:
                    # If just one is not present, we know the unit is included.
                    return True

                for key, value in rpm_package_req.items():
                    value_present = rpm_present.get(key)
                    if value_present != value:
                        _LOGGER.debug(
                            "%s: Not including unit as RPM %s has matching %r - expected %r got %r",
                            unit_name,
                            rpm_name,
                            key,
                            value,
                            value_present,
                        )
                        return True

        if not should_be_present:
            _LOGGER.debug("%s: Not including unit as all RPM are present in the runtime environment", unit_name)
            return False

        # Path to should be present.
        _LOGGER.debug("%s: all RPM package presence checks passed", unit_name)
        return True

    @classmethod
    def _prepare_justification_link(cls, entries: List[Dict[str, Any]]) -> None:
        """Prepare justification links before using them."""
        for entry in entries:
            link = entry.get("link")
            if link and not link.startswith(("https://", "http://")):
                entry["link"] = jl(link)

    @property
    def name(self) -> str:
        """Get name of the prescription instance."""
        name: str = self.prescription["name"]
        return name

    def pre_run(self) -> None:
        """Prepare this pipeline unit before running it."""
        self._prepare_justification_link(self.run_prescription.get("stack_info", []))
        self._stack_info_run = False
        super().pre_run()

    @staticmethod
    def _yield_should_include(unit_prescription: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
        """Yield for every entry stated in the match field."""
        match = unit_prescription.get("match", {})
        run = unit_prescription.get("run", {})
        if isinstance(match, list):
            for item in match:
                yield {
                    "package_name": item["package_version"].get("name") if item else None,
                    "match": item,
                    "run": run,
                }
        else:
            yield {"package_name": match["package_version"].get("name") if match else None, "match": match, "run": run}

    @staticmethod
    def _yield_should_include_with_state(unit_prescription: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
        """Yield for every entry stated in the match field."""
        match = unit_prescription.get("match", {})
        if isinstance(match, list):
            for item in match:
                match_resolved = item.get("state", {}).get("resolved_dependencies")
                yield {
                    # Return the first package name that should be matched to keep optimization for wrap calls.
                    "package_name": match_resolved[0].get("name") if match_resolved else None,
                    "match": item,
                    "run": unit_prescription["run"],
                }
        else:
            match_resolved = match.get("state", {}).get("resolved_dependencies") if match else None
            yield {
                "package_name": match_resolved[0].get("name") if match_resolved else None,
                "match": match,
                "run": unit_prescription["run"],
            }

    def _run_log(self) -> None:
        """Log message specified in the run prescription."""
        log = self.run_prescription.get("log")
        if log:
            _LOGGER.log(level=getattr(logging, log["type"]), msg=f"{self.name}: {log['message']}")

    def _run_stack_info(self) -> None:
        """Add stack info if any prescribed."""
        if self._stack_info_run:
            # Stack info already reported.
            return None

        self._stack_info_run = True
        stack_info = self.run_prescription.get("stack_info")
        if stack_info:
            self.context.stack_info.extend(stack_info)

    def _check_package_tuple_from_prescription(
        self, dependency_tuple: Tuple[str, str, str], dependency: Dict[str, str]
    ) -> bool:
        """Check if the given package version tuple matches with what was written in prescription."""
        develop = dependency.get("develop")
        if develop is not None:
            package_version = self.context.get_package_version(dependency_tuple, graceful=True)
            if not package_version:
                return False

            if package_version.develop != develop:
                return False

        if not self._index_url_check(dependency.get("index_url"), dependency_tuple[2]):
            return False

        version = dependency.get("version")
        if version is not None:
            specifier = SpecifierSet(version)
            if dependency_tuple[1] not in specifier:
                return False

        return True

    def _run_state(self, state: State) -> bool:
        """Check state match."""
        state_prescription = self.match_prescription.get("state")
        if not state_prescription:
            # Nothing to check.
            return True

        for resolved_dependency in state_prescription.get("resolved_dependencies", []):
            resolved = state.resolved_dependencies.get(resolved_dependency["name"])
            if not resolved:
                return False

            if not self._check_package_tuple_from_prescription(resolved, resolved_dependency):
                return False

        return True

    def _run_state_with_initiator(self, state: State, package_version: PackageVersion) -> bool:
        """Check state match respecting also initiator of the give package."""
        state_prescription = self.match_prescription.get("state")
        if not state_prescription:
            # Nothing to check.
            return True

        package_version_from = state_prescription.get("package_version_from") or []
        # XXX: we explicitly do not consider runtime environment as we expect to have it only one here.
        dependents = {
            i[0] for i in self.context.dependents.get(package_version.name, {}).get(package_version.to_tuple(), set())
        }

        for resolved_dependency in package_version_from:
            resolved = state.resolved_dependencies.get(resolved_dependency["name"])
            if not resolved:
                return False

            if not self._check_package_tuple_from_prescription(resolved, resolved_dependency):
                return False

            if resolved not in dependents:
                _LOGGER.debug(
                    "Package %r stated in package_version_from not a did not introduce package %r",
                    resolved,
                    package_version.to_tuple(),
                )
                return False

            dependents.discard(resolved)

        if dependents and not state_prescription.get("package_version_from_allow_other", False):
            for dependent in dependents:
                if dependent == state.resolved_dependencies.get(dependent[0]):
                    return False

        for resolved_dependency in state_prescription.get("resolved_dependencies", []):
            resolved = state.resolved_dependencies.get(resolved_dependency["name"])
            if not resolved:
                return False

            if not self._check_package_tuple_from_prescription(resolved, resolved_dependency):
                return False

        return True

    def _run_base(self) -> None:
        """Implement base routines for run part of the prescription."""
        self._run_log()
        self._run_stack_info()

        not_acceptable = self.run_prescription.get("not_acceptable")
        if not_acceptable:
            raise NotAcceptable(not_acceptable)

        eager_stop_pipeline = self.run_prescription.get("eager_stop_pipeline")
        if eager_stop_pipeline:
            raise EagerStopPipeline(eager_stop_pipeline)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test adding information about prescription release."""

from thoth.adviser.boots import PrescriptionReleaseBoot
from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.prescription import Prescription

from ..base import AdviserUnitTestCase


class TestPrescriptionReleaseBoot(AdviserUnitTestCase):
    """Test adding information about prescription release."""

    UNIT_TESTED = PrescriptionReleaseBoot

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.LATEST
        prescription_path = str(self.data_dir / "prescriptions")
        builder_context.prescription = Prescription.load(prescription_path)
        self.verify_multiple_should_include(builder_context)

    def test_no_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test not including this pipeline unit."""
        builder_context.prescription = None
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test including this pipeline unit."""
        prescription_path = str(self.data_dir / "prescriptions")
        builder_context.prescription = Prescription.load(prescription_path)
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

    def test_run(self, context: Context) -> None:
        """Test remapping UBI to RHEL."""
        assert not context.stack_info

        prescription_path = str(self.data_dir / "prescriptions")
        context.prescription = Prescription.load(prescription_path)

        unit = self.UNIT_TESTED()
        unit.pre_run()
        with self.UNIT_TESTED.assigned_context(context):
            unit.run()

        assert context.stack_info, "No stack info provided"
        assert len(context.stack_info) == 1
        assert context.stack_info[0]["type"] == "INFO"
        assert context.stack_info[0]["message"] == "Using prescriptions 'thoth' release '2021.06.15.dev'"
        assert self.verify_justification_schema(context.stack_info) is True



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test wrap that provides information derived from Python trove classifiers."""

import pytest
from typing import Dict
from typing import List
from typing import Optional
from thoth.common import get_justification_link as jl
from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.state import State
from thoth.adviser.wraps import TroveClassifiersWrap

from ..base import AdviserUnitTestCase


class TestTroveClassifiersWrap(AdviserUnitTestCase):
    """Test wrap that provides information derived from Python trove classifiers."""

    UNIT_TESTED = TroveClassifiersWrap

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        self.verify_multiple_should_include(builder_context)

    def test_run_classifiers(self, state: State, context: Context) -> None:
        """Test running and assigning all the classifiers."""
        trove_classifiers = [
            "Development Status:: 3 - Alpha",
            "Development Status:: 7 - Inactive",
            "Environment :: GPU :: NVIDIA CUDA",
            "Environment :: GPU :: NVIDIA CUDA :: 11.2",
            "Environment :: GPU :: NVIDIA CUDA :: 9.2",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.9",
        ]

        package_name = "tensorflow-gpu"
        package_version = "2.6.0"
        index_url = "https://pypi.org/simple"

        context.recommendation_type = RecommendationType.STABLE
        context.project.runtime_environment.cuda_version = "10.2"
        context.project.runtime_environment.python_version = "3.8"
        context.graph.should_receive("get_python_package_version_trove_classifiers_all").with_args(
            package_name=package_name,
            package_version=package_version,
            index_url=index_url,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
        ).and_return(trove_classifiers).once()

        state.justification.clear()

        unit = self.UNIT_TESTED()
        unit.pre_run()

        state.resolved_dependencies.clear()
        state.add_resolved_dependency((package_name, package_version, index_url))

        with unit.assigned_context(context):
            unit.run(state)

        assert state.justification == [
            {
                "link": jl("trove_cuda"),
                "message": "No CUDA specific trove classifier matching CUDA version used '10.2'",
                "package_name": package_name,
                "type": "WARNING",
            },
            {
                "link": jl("trove_unstable"),
                "message": "Development status stated in trove classifiers is 'alpha' which "
                "might be not suitable to be used with recommendation type "
                "'stable'",
                "package_name": "tensorflow-gpu",
                "type": "WARNING",
            },
            {
                "link": jl("trove_inactive"),
                "message": f"Inactive development status of {package_name!r} stated in trove " "classifiers",
                "package_name": package_name,
                "type": "WARNING",
            },
            {
                "link": jl("trove_py"),
                "message": "No Python specific trove classifier matching Python version used '3.8'",
                "package_name": package_name,
                "type": "WARNING",
            },
        ]

    @pytest.mark.parametrize(
        "cuda_version,trove_classifiers,justification",
        [
            (
                "10.2",
                ["Environment :: GPU :: NVIDIA CUDA :: 9.2", "Environment :: GPU :: NVIDIA CUDA :: 11.2"],
                [
                    {
                        "link": jl("trove_cuda"),
                        "message": "No CUDA specific trove classifier matching CUDA version used '10.2'",
                        "package_name": "tensorflow-gpu",
                        "type": "WARNING",
                    }
                ],
            ),
            (
                None,
                ["Environment :: GPU :: NVIDIA CUDA :: 11.0"],
                [],
            ),
            (
                "10.2",
                ["Environment :: GPU :: NVIDIA CUDA :: 9.2", "Environment :: GPU :: NVIDIA CUDA :: 10.2"],
                [],
            ),
        ],
    )
    def test_environment_gpu_cuda_version(
        self,
        cuda_version: Optional[str],
        trove_classifiers: List[str],
        context: Context,
        state: State,
        justification: List[Dict[str, str]],
    ) -> None:
        """Test adding justifications for environment GPU CUDA version."""
        package_name = "tensorflow-gpu"
        package_version = "2.6.0"
        index_url = "https://pypi.org/simple"

        context.project.runtime_environment.cuda_version = cuda_version
        context.graph.should_receive("get_python_package_version_trove_classifiers_all").with_args(
            package_name=package_name,
            package_version=package_version,
            index_url=index_url,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
        ).and_return(trove_classifiers).once()

        state.justification.clear()

        unit = self.UNIT_TESTED()
        unit.pre_run()

        state.resolved_dependencies.clear()
        state.add_resolved_dependency((package_name, package_version, index_url))

        with unit.assigned_context(context):
            unit.run(state)

        assert state.justification == justification

    @pytest.mark.parametrize(
        "development_status_classifiers,recommendation_type,justification",
        [
            ([" Development Status:: 1 - Planning  "], RecommendationType.TESTING, []),
            ([" Development Status:: 2 - Pre - Alpha "], RecommendationType.TESTING, []),
            (["Development Status:: 3 - Alpha"], RecommendationType.TESTING, []),
            (["Development Status:: 4 - Beta"], RecommendationType.TESTING, []),
            (
                ["Development Status:: 5 - Production / Stable"],
                RecommendationType.TESTING,
                [],
            ),
            (["Development Status:: 1 - Planning"], RecommendationType.LATEST, []),
            (["Development Status:: 2 - Pre - Alpha"], RecommendationType.LATEST, []),
            (["Development Status:: 3 - Alpha"], RecommendationType.LATEST, []),
            (["Development Status:: 4 - Beta"], RecommendationType.LATEST, []),
            (
                ["Development Status:: 5 - Production / Stable"],
                RecommendationType.LATEST,
                [],
            ),
            (
                ["Development Status:: 1 - Planning"],
                RecommendationType.STABLE,
                [
                    {
                        "link": jl("trove_unstable"),
                        "message": "Development status stated in trove classifiers is 'planning' "
                        "which might be not suitable to be used with recommendation type "
                        "'stable'",
                        "package_name": "thoth-common",
                        "type": "WARNING",
                    }
                ],
            ),
            (
                ["Development Status:: 2 - Pre - Alpha"],
                RecommendationType.PERFORMANCE,
                [
                    {
                        "link": jl("trove_unstable"),
                        "message": "Development status stated in trove classifiers is 'pre-alpha' "
                        "which might be not suitable to be used with recommendation type "
                        "'performance'",
                        "package_name": "thoth-common",
                        "type": "WARNING",
                    }
                ],
            ),
            (
                ["Development Status:: 3 - Alpha"],
                RecommendationType.STABLE,
                [
                    {
                        "link": jl("trove_unstable"),
                        "message": "Development status stated in trove classifiers is 'alpha' "
                        "which might be not suitable to be used with recommendation type "
                        "'stable'",
                        "package_name": "thoth-common",
                        "type": "WARNING",
                    }
                ],
            ),
            (
                ["Development Status:: 4 - Beta"],
                RecommendationType.SECURITY,
                [
                    {
                        "link": jl("trove_unstable"),
                        "message": "Development status stated in trove classifiers is 'beta' "
                        "which might be not suitable to be used with recommendation type "
                        "'security'",
                        "package_name": "thoth-common",
                        "type": "WARNING",
                    }
                ],
            ),
            (
                ["Development Status:: 5 - Production / Stable"],
                RecommendationType.STABLE,
                [],
            ),
            (
                ["Development Status:: 6 - Mature"],
                RecommendationType.STABLE,
                [],
            ),
        ],
    )
    def test_development_status(
        self,
        development_status_classifiers: List[str],
        recommendation_type: RecommendationType,
        justification: List[Dict[str, str]],
        state: State,
        context: Context,
    ) -> None:
        """Test adding justifications for development status."""
        package_name = "thoth-common"
        package_version = "0.26.0"
        index_url = "https://pypi.org/simple"

        context.graph.should_receive("get_python_package_version_trove_classifiers_all").with_args(
            package_name=package_name,
            package_version=package_version,
            index_url=index_url,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
        ).and_return(development_status_classifiers).once()
        context.recommendation_type = recommendation_type

        state.justification.clear()

        unit = self.UNIT_TESTED()
        unit.pre_run()

        state.resolved_dependencies.clear()
        state.add_resolved_dependency((package_name, package_version, index_url))

        with unit.assigned_context(context):
            unit.run(state)

        assert state.justification == justification

    @pytest.mark.parametrize("recommendation_type", list(RecommendationType.__members__.values()))
    def test_development_status_inactive(
        self,
        recommendation_type: RecommendationType,
        state: State,
        context: Context,
    ) -> None:
        """Test adding justifications for inactive development status."""
        package_name = "argparse"
        package_version = "1.4.0"
        index_url = "https://pypi.org/simple"

        development_status_classifiers = ["Development Status:: 7 - Inactive"]

        context.recommendation_type = recommendation_type
        context.graph.should_receive("get_python_package_version_trove_classifiers_all").with_args(
            package_name=package_name,
            package_version=package_version,
            index_url=index_url,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
        ).and_return(development_status_classifiers).once()

        state.justification.clear()

        unit = self.UNIT_TESTED()
        unit.pre_run()

        state.resolved_dependencies.clear()
        state.add_resolved_dependency((package_name, package_version, index_url))

        with unit.assigned_context(context):
            unit.run(state)

        assert state.justification == [
            {
                "link": jl("trove_inactive"),
                "message": f"Inactive development status of {package_name!r} stated in trove classifiers",
                "package_name": package_name,
                "type": "WARNING",
            }
        ]

    @pytest.mark.parametrize(
        "python_version,justification",
        [
            (
                "3.6",
                [
                    {
                        "link": jl("trove_py"),
                        "message": "No Python specific trove classifier matching Python version used '3.6'",
                        "package_name": "tensorflow",
                        "type": "WARNING",
                    }
                ],
            ),
            (
                "3.8",
                [],
            ),
            (
                None,
                [],
            ),
        ],
    )
    def test_programming_language_python(
        self, python_version: Optional[str], justification: List[Dict[str, str]], state: State, context: Context
    ) -> None:
        """Test adding justifications for Python programming language."""
        trove_classifiers = [
            "  Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9  ",
            "Programming Language :: Python :: 3.10",
        ]

        package_name = "tensorflow"
        package_version = "2.6.0"
        index_url = "https://pypi.org/simple"

        context.project.runtime_environment.python_version = python_version
        context.graph.should_receive("get_python_package_version_trove_classifiers_all").with_args(
            package_name=package_name,
            package_version=package_version,
            index_url=index_url,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
        ).and_return(trove_classifiers).once()

        state.justification.clear()

        unit = self.UNIT_TESTED()
        unit.pre_run()

        state.resolved_dependencies.clear()
        state.add_resolved_dependency((package_name, package_version, index_url))

        with unit.assigned_context(context):
            unit.run(state)

        assert state.justification == justification

    def test_unknown_trove_classifiers(self, state: State, context: Context) -> None:
        """Test not adding justifications for unknown trove classifiers."""
        trove_classifiers = [
            "2323938",
            " Language :: Python :: 3.9  ",
            " Foo: bar",
            "Environment :: GPU :: NVIDIA CUDA :: 9.2.32",
        ]

        package_name = "cowsay"
        package_version = "1.0.0"
        index_url = "https://pypi.org/simple"

        context.project.runtime_environment.python_version = "3.9"
        context.project.runtime_environment.cuda_version = "10.0"
        context.graph.should_receive("get_python_package_version_trove_classifiers_all").with_args(
            package_name=package_name,
            package_version=package_version,
            index_url=index_url,
            os_name=context.project.runtime_environment.operating_system.name,
            os_version=context.project.runtime_environment.operating_system.version,
            python_version=context.project.runtime_environment.python_version,
        ).and_return(trove_classifiers).once()

        state.justification.clear()

        unit = self.UNIT_TESTED()
        unit.pre_run()

        state.resolved_dependencies.clear()
        state.add_resolved_dependency((package_name, package_version, index_url))

        with unit.assigned_context(context):
            unit.run(state)

        assert state.justification == []



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of hill climbing in the state space."""

import logging

import attr
from typing import List
from typing import Tuple

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties

from ..predictor import Predictor
from ..state import State
from ..exceptions import NoHistoryKept


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class HillClimbing(Predictor):
    """Implementation of hill climbing in the state space."""

    _history = attr.ib(type=List[Tuple[float, int]], default=attr.Factory(list), init=False)

    def run(self) -> Tuple[State, Tuple[str, str, str]]:
        """Get top state from the beam for the next resolution round."""
        state = self.context.beam.max()

        if self.keep_history:
            self._history.append((state.score, self.context.accepted_final_states_count))

        return state, state.get_first_unresolved_dependency()

    def pre_run(self) -> None:
        """Initialize before the actual hill climbing run."""
        self._history = []

    def plot(self) -> matplotlib.figure.Figure:
        """Plot score of the highest rated stack during hill climbing."""
        if not self._history:
            raise NoHistoryKept("No history datapoints kept")

        x = [i for i in range(len(self._history))]
        y1 = [i[0] for i in self._history]
        y2 = [i[1] for i in self._history]

        fig, host = plt.subplots()
        fig.subplots_adjust(right=0.75)

        par1 = host.twinx()

        par1.spines["right"].set_position(("axes", 1.10))
        self._make_patch_spines_invisible(par1)

        par1.spines["right"].set_visible(True)
        host.spines["right"].set_visible(False)
        host.spines["top"].set_visible(False)

        (p1,) = host.plot(x, y1, ",g", label="Score of the expanded state")
        (p2,) = par1.plot(x, y2, ",y", label="Number of products conducted")

        host.set_xlabel("iteration")
        host.set_ylabel("score")
        par1.set_ylabel("product count")

        host.yaxis.label.set_color(p1.get_color())
        par1.yaxis.label.set_color(p2.get_color())

        tkw = dict(size=4, width=1.5)
        host.tick_params(axis="y", colors=p1.get_color(), **tkw)
        host.tick_params(axis="x", **tkw)
        par1.tick_params(axis="y", colors=p2.get_color(), **tkw)

        font_prop = FontProperties()
        font_prop.set_size("medium")
        fig.legend(
            loc="upper center",
            bbox_to_anchor=(0.50, 1.00),
            ncol=2,
            fancybox=True,
            shadow=True,
            prop=font_prop,
        )
        return fig



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test removing pre-releases in direct dependencies."""

import flexmock
import pytest

from thoth.adviser.sieves import CutPreReleasesSieve
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.python import Source
from thoth.python import PackageVersion
from thoth.python import Project

from ..base import AdviserUnitTestCase


class TestCutPreReleasesSieve(AdviserUnitTestCase):
    """Test removing dependencies based on pre-releases configuration."""

    UNIT_TESTED = CutPreReleasesSieve

    _CASE_DISALLOWED_PIPFILE = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[[source]]
url = "https://tensorflow.pypi.thoth-station.ninja/index/os/fedora/30/jemalloc/simple/"
verify_ssl = true
name = "aicoe"

[packages]
tensorflow = "*"

[pipenv]
allow_prereleases = false
"""

    _CASE_ALLOWED_PIPFILE = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[[source]]
url = "https://tensorflow.pypi.thoth-station.ninja/index/os/fedora/30/jemalloc/simple/"
verify_ssl = true
name = "aicoe"

[packages]
tensorflow = "*"

[pipenv]
allow_prereleases = true
"""

    _CASE_SELECTIVE_PRERELEASES_ALLOWED_PIPFILE = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"

[pipenv]
allow_prereleases = true

[thoth.allow_prereleases]
tensorflow = true
"""

    @pytest.mark.parametrize(
        "recommendation_type",
        [
            RecommendationType.STABLE,
            RecommendationType.PERFORMANCE,
            RecommendationType.SECURITY,
            RecommendationType.LATEST,
            RecommendationType.TESTING,
        ],
    )
    def test_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
    ) -> None:
        """Test including this pipeline unit."""
        builder_context.recommendation_type = recommendation_type
        builder_context.project = Project.from_strings(self._CASE_DISALLOWED_PIPFILE)

        assert builder_context.is_adviser_pipeline()
        assert list(CutPreReleasesSieve.should_include(builder_context)) == [{}]

    @pytest.mark.parametrize(
        "recommendation_type",
        [
            RecommendationType.STABLE,
            RecommendationType.PERFORMANCE,
            RecommendationType.SECURITY,
            RecommendationType.LATEST,
            RecommendationType.TESTING,
        ],
    )
    def test_not_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
    ) -> None:
        """Test not including this pipeline unit."""
        builder_context.recommendation_type = recommendation_type
        builder_context.project = Project.from_strings(self._CASE_ALLOWED_PIPFILE)

        assert builder_context.is_adviser_pipeline()
        assert list(CutPreReleasesSieve.should_include(builder_context)) == []

    def test_not_include_thoth_prereleases_allowed(self, builder_context: PipelineBuilderContext) -> None:
        """Test not including this pipeline unit."""
        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.project = Project.from_strings(self._CASE_SELECTIVE_PRERELEASES_ALLOWED_PIPFILE)

        assert builder_context.is_adviser_pipeline()
        assert list(CutPreReleasesSieve.should_include(builder_context)) == []

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.project = Project.from_strings(self._CASE_DISALLOWED_PIPFILE)
        self.verify_multiple_should_include(builder_context)

    def test_remove_pre_releases_disallowed_noop(self) -> None:
        """Test removing dependencies not hitting limit causes a noop."""
        tf_2_0_0 = PackageVersion(
            name="tensorflow",
            version="==2.0.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_DISALLOWED_PIPFILE))
        with CutPreReleasesSieve.assigned_context(context):
            sieve = CutPreReleasesSieve()
            assert list(sieve.run(p for p in [tf_2_0_0])) == [tf_2_0_0]

    def test_pre_releases_disallowed_removal(self) -> None:
        """Test no removals if pre-releases are allowed."""
        tf_2_0_0rc0 = PackageVersion(
            name="tensorflow",
            version="==2.0.0rc0",
            index=Source("https://tensorflow.pypi.thoth-station.ninja/index/os/fedora/30/jemalloc/simple/"),
            develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_DISALLOWED_PIPFILE))
        with CutPreReleasesSieve.assigned_context(context):
            sieve = CutPreReleasesSieve()
            assert list(sieve.run(p for p in [tf_2_0_0rc0])) == []



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A prescription step implementing adding a package to a dependency graph."""

import logging

import attr
from thoth.adviser.state import State
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages.exceptions import NotFoundError
from voluptuous import Any as SchemaAny
from voluptuous import Schema
from voluptuous import Required

from .step import StepPrescription
from .schema import PRESCRIPTION_ADD_PACKAGE_STEP_RUN_SCHEMA
from .schema import PRESCRIPTION_ADD_PACKAGE_STEP_MATCH_ENTRY_SCHEMA

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class AddPackageStepPrescription(StepPrescription):
    """Add package prescription step unit implementation."""

    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): SchemaAny(str, None),
            Required("match"): PRESCRIPTION_ADD_PACKAGE_STEP_MATCH_ENTRY_SCHEMA,
            Required("multi_package_resolution"): bool,
            Required("run"): SchemaAny(PRESCRIPTION_ADD_PACKAGE_STEP_RUN_SCHEMA, None),
        }
    )

    def run(self, state: State, package_version: PackageVersion) -> None:
        """Run main entry-point for steps to skip packages."""
        if not self._index_url_check(self._index_url, package_version.index.url):
            return None

        if self._specifier and package_version.locked_version not in self._specifier:
            return None

        if self._develop is not None and package_version.develop != self._develop:
            return None

        if not self._run_state_with_initiator(state, package_version):
            return None

        add_package_version = self.run_prescription["package_version"]
        add_package_version_name = add_package_version["name"]
        add_package_version_version = add_package_version["locked_version"][2:]
        add_package_version_index_url = add_package_version["index_url"]
        add_package_version_develop = add_package_version["develop"]

        add_package_version_tuple = (
            add_package_version_name,
            add_package_version_version,
            add_package_version_index_url,
        )

        resolved = state.resolved_dependencies.get(add_package_version_name)
        if resolved:
            if resolved == add_package_version_tuple:
                _LOGGER.debug(
                    "%s: Not adding package %r as it is already in the resolved listing",
                    self.get_unit_name(),
                    add_package_version_tuple,
                )
            else:
                _LOGGER.debug(
                    "%s: Not adding package %r as another package %r is already present in the resolved listing",
                    self.get_unit_name(),
                    add_package_version_tuple,
                    resolved,
                )

            return None

        runtime_env = self.context.project.runtime_environment
        py_ver = runtime_env.python_version.replace(".", "")
        # XXX: this could be moved to thoth-common
        solver_name = f"solver-{runtime_env.operating_system.name}-{runtime_env.operating_system.version}-py{py_ver}"
        if not self.context.graph.python_package_version_exists(
            add_package_version_name,
            add_package_version_version,
            add_package_version_index_url,
            solver_name=solver_name,
        ):
            _LOGGER.debug(
                "%s: Not adding package %r as the given package was not solved by %r",
                self.get_unit_name(),
                add_package_version_tuple,
                solver_name,
            )
            return None

        try:
            if not self.context.graph.is_python_package_index_enabled(add_package_version_index_url):
                _LOGGER.debug(
                    "%s: Not adding package %r as index %r is not enabled",
                    self.get_unit_name(),
                    add_package_version_tuple,
                    add_package_version_index_url,
                )
                return None
        except NotFoundError:
            _LOGGER.debug(
                "%s: Not adding package %r as index %r is not known to the resolver",
                self.get_unit_name(),
                add_package_version_tuple,
                add_package_version_index_url,
            )
            return None

        self._run_base()

        pv = PackageVersion(
            name=add_package_version_name,
            version=add_package_version["locked_version"],
            index=Source(add_package_version_index_url),
            develop=add_package_version_develop,
        )
        self.context.register_package_version(pv)
        state.add_unresolved_dependency(add_package_version_tuple)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to check if GPU specific recommendations should be done."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr

from thoth.common import get_justification_link as jl

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class GPUBoot(Boot):
    """A boot to check if GPU specific recommendations should be done."""

    _MESSAGE_NO_GPU = "CUDA version provided but no GPU model available, turning off GPU specific recommendations"
    _MESSAGE_NO_CUDA = "GPU model provided but no CUDA version available, turning off GPU specific recommendations"
    _JUSTIFICATION_LINK = jl("no_gpu")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, always."""
        if not builder_context.is_included(cls) and (
            builder_context.project.runtime_environment.cuda_version
            or builder_context.project.runtime_environment.hardware.gpu_model
        ):
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Check if GPU enabled recommendations should be done."""
        if (
            self.context.project.runtime_environment.cuda_version
            and not self.context.project.runtime_environment.hardware.gpu_model
        ):
            _LOGGER.warning(self._MESSAGE_NO_GPU)
            self.context.stack_info.append(
                {
                    "type": "WARNING",
                    "message": self._MESSAGE_NO_GPU,
                    "link": self._JUSTIFICATION_LINK,
                }
            )
            self.context.project.runtime_environment.cuda_version = None
        elif (
            not self.context.project.runtime_environment.cuda_version
            and self.context.project.runtime_environment.hardware.gpu_model
        ):
            _LOGGER.warning(self._MESSAGE_NO_CUDA)
            self.context.stack_info.append(
                {
                    "type": "WARNING",
                    "message": self._MESSAGE_NO_CUDA,
                    "link": self._JUSTIFICATION_LINK,
                }
            )
            self.context.project.runtime_environment.hardware.gpu_model = None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of generic v1 prescription unit handling."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import flexmock
import pytest

from thoth.adviser.enums import RecommendationType
from thoth.adviser.enums import DecisionType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.prescription.v1 import UnitPrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_UNIT_SHOULD_INCLUDE_RUNTIME_ENVIRONMENTS_SCHEMA
from thoth.common import get_justification_link as jl
from thoth.common import RuntimeEnvironment

from ...base import AdviserTestCase


class TestUnitPrescription(AdviserTestCase):
    """Test implementation of generic v1 prescription unit handling."""

    _EXAMPLE_RPM_PACKAGES_PRESENT = [
        {
            "arch": "x86_64",
            "epoch": None,
            "package_identifier": "zlib-1.2.11-16.2.el8_3.x86_64",
            "package_name": "zlib",
            "package_version": "1.2.11",
            "release": "16.2.el8_3",
            "src": False,
        },
        {
            "arch": "x86_64",
            "epoch": None,
            "package_identifier": "glib2-2.56.4-8.el8.x86_64",
            "package_name": "glib2",
            "package_version": "2.56.4",
            "release": "8.el8",
            "src": False,
        },
        {
            "arch": "x86_64",
            "epoch": None,
            "package_identifier": "gcc-8.3.1-5.1.el8.x86_64",
            "package_name": "gcc",
            "package_version": "8.3.1",
            "release": "5.1.el8",
            "src": False,
        },
        {
            "arch": "x86_64",
            "epoch": "1",
            "package_identifier": "findutils-1:4.6.0-20.el8.x86_64",
            "package_name": "findutils",
            "package_version": "4.6.0",
            "release": "20.el8",
            "src": False,
        },
    ]

    _EXAMPLE_PYTHON_PACKAGES_PRESENT = [
        {
            "location": "/usr/local/lib/python3.8/site-packages",
            "package_name": "pip",
            "package_version": "20.3.3",
        },
        {
            "location": "/usr/lib64/python3.6/site-packages",
            "package_name": "dbus-python",
            "package_version": "1.2.4",
        },
        {
            "location": "/usr/lib64/python3.6/site-packages",
            "package_name": "gpg",
            "package_version": "1.13.1",
        },
        {
            "location": "/usr/lib64/python3.6/site-packages",
            "package_name": "rpm",
            "package_version": "4.14.3",
        },
        {
            "location": "/usr/lib/python3.6/site-packages",
            "package_name": "requests",
            "package_version": "2.20.0",
        },
        {
            "location": "/opt/app-root/lib/python3.8/site-packages",
            "package_name": "wheel",
            "package_version": "0.36.2",
        },
        {
            "location": "/opt/app-root/lib/python3.8/site-packages",
            "package_name": "yarl",
            "package_version": "1.6.3",
        },
        {
            "location": "/opt/app-root/lib/python3.8/site-packages",
            "package_name": "yaspin",
            "package_version": "2.0.0",
        },
    ]

    def test_prepare_justification_link(self) -> None:
        """Test preparing links to justifications."""
        justification = [
            {"type": "INFO", "message": "Some info message", "link": "s2i_thoth"},
            {"type": "WARNING", "message": "Some warning message", "link": "https://thoth-station.ninja"},
            {"type": "ERROR", "message": "Some error message", "link": "https://thoth-station.ninja"},
        ]

        flexmock(UnitPrescription, run_prescription=justification)
        UnitPrescription._prepare_justification_link(justification)

        assert justification == [
            {"type": "INFO", "message": "Some info message", "link": jl("s2i_thoth")},
            {"type": "WARNING", "message": "Some warning message", "link": "https://thoth-station.ninja"},
            {"type": "ERROR", "message": "Some error message", "link": "https://thoth-station.ninja"},
        ]

    def test_should_include_dependencies(self, caplog, builder_context: PipelineBuilderContext) -> None:
        """Test including a pipeline unit based on dependencies."""
        assert builder_context.is_adviser_pipeline()

        should_include_dict = {"adviser_pipeline": True, "dependencies": {"boots": ["Foo"]}}
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {"name": "SomeUnitName", "should_include": should_include_dict}
        builder_context.should_receive("get_included_boot_names").and_yield("Foo")
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False)
        assert UnitPrescription._should_include_base(builder_context) is True

        should_include_dict = {"adviser_pipeline": True, "dependencies": {"boots": ["Foo"]}}
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {"name": "SomeUnitName", "should_include": should_include_dict}
        builder_context.should_receive("get_included_boot_names").and_yield()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False)
        assert UnitPrescription._should_include_base(builder_context) is False

        should_include_dict = {
            "adviser_pipeline": True,
            "dependencies": {
                "boots": ["BootUnit1", "BootUnit2"],
                "pseudonyms": ["PseudonymUnit1"],
                "sieves": ["SieveUnit1"],
                "steps": ["StepUnit1"],
                "strides": ["StrideUnit1"],
                "wraps": ["WrapUnit1"],
            },
        }
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {"name": "SomeUnitName", "should_include": should_include_dict}
        builder_context.should_receive("get_included_boot_names").and_yield("BootUnit1", "BootUnit2")
        builder_context.should_receive("get_included_pseudonym_names").and_yield("PseudonymUnit1", "PseudonymUnit2")
        builder_context.should_receive("get_included_sieve_names").and_yield("SieveUnit1", "SieveUnit2")
        builder_context.should_receive("get_included_step_names").and_yield("StepUnit0", "StepUnit1")
        builder_context.should_receive("get_included_stride_names").and_yield("StrideUnit1")
        builder_context.should_receive("get_included_wrap_names").and_yield("WrapUnit1")
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False)
        assert UnitPrescription._should_include_base(builder_context) is True

    @pytest.mark.parametrize(
        "recommendation_type,allowed_recommendation_types,include",
        [
            (RecommendationType.LATEST, None, True),
            (RecommendationType.LATEST, ["latest"], True),
            (RecommendationType.STABLE, ["latest", "performance"], False),
            (RecommendationType.STABLE, ["latest", "performance", "stable"], True),
        ],
    )
    def test_should_include_recommendation_type(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
        allowed_recommendation_types: Optional[List[str]],
        include: bool,
    ) -> None:
        """Test including a pipeline unit based on recommendation type."""
        should_include_dict = {"adviser_pipeline": True}
        if allowed_recommendation_types is not None:
            should_include_dict["recommendation_types"] = allowed_recommendation_types

        builder_context.recommendation_type = recommendation_type
        builder_context.decision_type = None

        assert builder_context.is_adviser_pipeline()
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {
            "name": "SomePrescriptionUnitName",
            "should_include": should_include_dict,
        }
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) == include

    @pytest.mark.parametrize(
        "decision_type,allowed_decision_types,include",
        [
            (DecisionType.RANDOM, None, True),
            (DecisionType.RANDOM, ["random"], True),
            (DecisionType.ALL, ["random"], False),
            (DecisionType.ALL, ["random", "all"], True),
        ],
    )
    def test_should_include_decision_type(
        self,
        builder_context: PipelineBuilderContext,
        decision_type: DecisionType,
        allowed_decision_types: List[str],
        include: bool,
    ) -> None:
        """Test including a pipeline unit based on decision type."""
        should_include_dict = {"dependency_monkey_pipeline": True}
        if allowed_decision_types is not None:
            should_include_dict["decision_types"] = allowed_decision_types

        builder_context.recommendation_type = None
        builder_context.decision_type = decision_type

        assert builder_context.is_dependency_monkey_pipeline()
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {
            "name": "SomePrescriptionUnitName",
            "should_include": should_include_dict,
        }
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) == include

    def test_should_include_pipeline(self, builder_context: PipelineBuilderContext) -> None:
        """Test including based on pipeline type."""
        builder_context.recommendation_type = None
        builder_context.decision_type = DecisionType.ALL
        assert builder_context.is_dependency_monkey_pipeline()

        should_include_dict = {"dependency_monkey_pipeline": True}
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {
            "name": "SomePrescriptionUnitName",
            "should_include": should_include_dict,
        }
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False)
        # builder context has decision type ALL and prescription has dependency monkey pipeline configured
        assert UnitPrescription._should_include_base(builder_context) is True
        UnitPrescription.SHOULD_INCLUDE_CACHE.clear()

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None
        assert builder_context.is_adviser_pipeline()

        # builder context has recommendation type LATEST and prescription has dependency monkey pipeline configured
        assert UnitPrescription._should_include_base(builder_context) is False
        UnitPrescription.SHOULD_INCLUDE_CACHE.clear()

        should_include_dict = {"adviser_pipeline": True}
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {
            "name": "SomePrescriptionUnitName",
            "should_include": should_include_dict,
        }
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False)
        # builder context has recommendation type LATEST and prescription has adviser pipeline configured
        assert UnitPrescription._should_include_base(builder_context) is True
        UnitPrescription.SHOULD_INCLUDE_CACHE.clear()

        builder_context.recommendation_type = None
        builder_context.decision_type = DecisionType.ALL
        assert builder_context.is_dependency_monkey_pipeline()

        # builder context has decision type ALL and prescription has adviser pipeline configured
        assert UnitPrescription._should_include_base(builder_context) is False
        UnitPrescription.SHOULD_INCLUDE_CACHE.clear()

    @pytest.mark.parametrize("adviser_pipeline", [True, False])
    def test_should_include_times(self, builder_context: PipelineBuilderContext, adviser_pipeline: bool) -> None:
        """Test including a pipeline unit based on `times`."""
        if adviser_pipeline:
            builder_context.recommendation_type = RecommendationType.LATEST
            builder_context.decision_type = None
            assert builder_context.is_adviser_pipeline()
            assert not builder_context.is_dependency_monkey_pipeline()
        else:
            builder_context.recommendation_type = None
            builder_context.decision_type = DecisionType.ALL
            assert builder_context.is_dependency_monkey_pipeline()
            assert not builder_context.is_adviser_pipeline()

        should_include_dict = {"times": 0, "adviser_pipeline": True, "dependency_monkey_pipeline": True}
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {
            "name": "Foo",
            "should_include": should_include_dict,
        }
        assert UnitPrescription._should_include_base(builder_context) is False

        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).and_return(
            True
        ).twice()

        should_include_dict = {"times": 1, "adviser_pipeline": True, "dependency_monkey_pipeline": True}
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {
            "name": "Foo",
            "should_include": should_include_dict,
        }
        assert UnitPrescription._should_include_base(builder_context) is True

        should_include_dict = {"times": 1, "adviser_pipeline": True, "dependency_monkey_pipeline": True}
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include_dict)
        UnitPrescription._PRESCRIPTION = {
            "name": "Foo",
            "should_include": should_include_dict,
        }
        assert UnitPrescription._should_include_base(builder_context) is False

    @pytest.mark.parametrize(
        "prescription_runtime_environments,used_runtime_environment_dict,include",
        [
            # Python version.
            ({"python_version": ">=3.6,<=3.9"}, {"python_version": "3.6"}, True),
            ({"python_version": ">=3.6,<=3.9"}, {}, False),
            ({"python_version": "!=3.6,!=3.9"}, {}, False),
            ({"python_version": "==3.6"}, {"python_version": "3.9"}, False),
            ({"python_version": ">=0.0"}, {}, False),
            # Operating system.
            ({"operating_systems": [{"name": "rhel", "version": "8"}]}, {}, False),
            (
                {"operating_systems": [{"name": "rhel", "version": "8"}, {"name": "fedora", "version": "33"}]},
                {"operating_system": {"name": "fedora", "version": "8"}},
                False,
            ),
            (
                {"operating_systems": [{"name": "rhel", "version": "8"}, {"name": "fedora", "version": "33"}]},
                {"operating_system": {"name": "rhel", "version": "8"}},
                True,
            ),
            (
                {"operating_systems": [{"name": "rhel"}]},
                {"operating_system": {"name": "rhel", "version": "8"}},
                False,
            ),
            # CPU/GPU.
            (
                {"hardware": [{"cpu_families": [1, 2, 3], "cpu_models": [9, 8, 7], "gpu_models": [None, "Some"]}]},
                {"hardware": {"cpu_family": 1, "cpu_model": 7}},
                True,
            ),
            (
                {"hardware": [{"cpu_families": {"not": [1]}}]},
                {"hardware": {"cpu_family": 1}},
                False,
            ),
            (
                {"hardware": [{"cpu_models": {"not": [None]}}]},
                {"hardware": {"cpu_model": 1}},
                True,
            ),
            (
                {"hardware": [{"gpu_models": {"not": ["Bar"]}}]},
                {"hardware": {"gpu_model": "Foo"}},
                True,
            ),
            (
                {"hardware": [{"cpu_families": [1, 2, 3], "cpu_models": [9, 8, 7]}]},
                {"hardware": {"cpu_family": 2, "cpu_model": 6}},
                False,
            ),
            (
                {
                    "hardware": [
                        {"cpu_families": [1, 2, 3], "cpu_models": [9, 8, 7]},
                        {"cpu_families": [3], "cpu_models": [8, 7], "gpu_models": ["Some"]},
                    ]
                },
                {"hardware": {"cpu_family": 3, "cpu_model": 8, "gpu_model": "Some"}},
                True,
            ),
            # CUDA versions.
            ({"cuda_version": "<=9.0,>=8.1"}, {"cuda_version": "9.0"}, True),
            ({"cuda_version": "!=9.0,!=8.1"}, {"cuda_version": "9.0"}, False),
            ({"cuda_version": "<=9.0,>=8.1"}, {}, False),
            ({"cuda_version": "==8.1"}, {"cuda_version": "9.0"}, False),
            # platforms.
            ({"platforms": ["linux-x86_64", "linux-i586"]}, {"platform": "linux-x86_64"}, True),
            ({"platforms": ["linux-x86_64", "linux-i586"]}, {}, False),
            ({"platforms": ["linux-i586"]}, {"platform": "linux-x86_64"}, False),
            ({"platforms": ["linux-i586", None]}, {}, True),
            ({"platforms": {"not": ["linux-i586", None]}}, {}, False),
            # openblas_version.
            ({"openblas_version": "<=0.3.1,>=0.2.0"}, {"openblas_version": "0.3.1"}, True),
            ({"openblas_version": "<=0.3.1,>=0.2.0"}, {}, False),
            ({"openblas_version": "==0.2.0"}, {"openblas_version": "0.3.1"}, False),
            ({"openblas_version": "!=0.2.0"}, {}, False),
            # openmpi_version.
            ({"openmpi_version": "<=3.1,>=2.0"}, {"openmpi_version": "3.1"}, True),
            ({"openmpi_version": "<=3.1,>=2.0"}, {}, False),
            ({"openmpi_version": "==2.0"}, {"openmpi_version": "3.1"}, False),
            ({"openmpi_version": ">=0.0"}, {}, False),
            # cudnn_version.
            ({"cudnn_version": "<=8.1,>=8.0"}, {"cudnn_version": "8.1"}, True),
            ({"cudnn_version": "<=8.1,>=8.0"}, {}, False),
            ({"cudnn_version": "==8.0"}, {"cudnn_version": "8.1"}, False),
            ({"cudnn_version": "!=8.0"}, {}, False),
            # mkl_version.
            ({"mkl_version": "<=2021.1.1,>=2019.0"}, {"mkl_version": "2021.1.1"}, True),
            ({"mkl_version": "<=2021.1.1,>=2019.0"}, {}, False),
            ({"mkl_version": "==2019.0"}, {"mkl_version": "2021.1.1"}, False),
            ({"mkl_version": "!=2019.0"}, {"mkl_version": "2021.1.1"}, True),
            # base images.
            (
                {
                    "base_images": [
                        "quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0",
                        "quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0",
                    ]
                },
                {"base_image": "quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0"},
                True,
            ),
            (
                {
                    "base_images": [
                        "quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0",
                        "quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0",
                    ]
                },
                {},
                False,
            ),
            (
                {"base_images": ["quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0"]},
                {"base_image": "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0"},
                False,
            ),
            ({"base_images": ["quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0", None]}, {}, True),
            (
                {"base_images": {"not": ["quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0"]}},
                {"base_image": "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0"},
                True,
            ),
            (
                {
                    "base_images": {
                        "not": [
                            "quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0",
                            "quay.io/thoth-station/s2i-thoth-ubi8-py36:v1.0.0",
                        ]
                    }
                },
                {},
                True,
            ),
            # Configuration combination.
            (
                {
                    "python_version": ">=3.6,<=3.9",
                    "cuda_version": "==9.0",
                    "cudnn_version": "<=3.1,>=3.0",
                    "base_images": [None, "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0"],
                    "platforms": ["linux-x86_64"],
                    "operating_systems": [{"name": "rhel", "version": "8.0"}, {"name": "fedora", "version": "30"}],
                    "hardware": [{"cpu_families": [1, 2, 3], "cpu_models": [9, 8, 7], "gpu_models": [None, "Some"]}],
                },
                {
                    "python_version": "3.6",
                    "cuda_version": "9.0",
                    "cudnn_version": "3.1",
                    "base_image": "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0",
                    "platform": "linux-x86_64",
                    "operating_system": {"name": "rhel", "version": "8.0"},
                    "hardware": {"cpu_family": 1, "cpu_model": 7},
                },
                True,
            ),
            (
                {
                    "python_version": ">=3.6,<=3.9",
                    "cuda_version": "==9.0",
                    "cudnn_version": "<=3.1,>=3.0",
                    "base_images": [None, "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0"],
                    "platforms": ["linux-x86_64"],
                    "operating_systems": [{"name": "rhel", "version": "8.0"}, {"name": "fedora", "version": "30"}],
                    "hardware": [
                        # CPU model does not match.
                        {"cpu_families": [1, 2, 3], "cpu_models": [9, 8], "gpu_models": [None, "Some"]}
                    ],
                },
                {
                    "python_version": "3.6",
                    "cuda_version": "9.0",
                    "cudnn_version": "3.1",
                    "base_image": "quay.io/thoth-station/s2i-thoth-ubi8-py38:v1.0.0",
                    "platform": "linux-x86_64",
                    "operating_system": {"name": "rhel", "version": "8.0"},
                    "hardware": {"cpu_family": 1, "cpu_model": 7},
                },
                False,
            ),
        ],
    )
    def test_should_include_runtime_environment(
        self,
        builder_context: PipelineBuilderContext,
        prescription_runtime_environments: Dict[str, Any],
        used_runtime_environment_dict: Dict[str, Any],
        include: bool,
    ) -> None:
        """Test parsing and including the given should include entry."""
        PRESCRIPTION_UNIT_SHOULD_INCLUDE_RUNTIME_ENVIRONMENTS_SCHEMA(prescription_runtime_environments)
        builder_context.project.runtime_environment = RuntimeEnvironment.from_dict(used_runtime_environment_dict)
        UnitPrescription._PRESCRIPTION = {
            "name": "Bar",
            "should_include": {
                "adviser_pipeline": True,
                "runtime_environments": prescription_runtime_environments,
            },
        }
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()

        assert builder_context.is_adviser_pipeline()
        assert UnitPrescription._should_include_base(builder_context) == include

    @pytest.mark.parametrize(
        "library_usage_expected,library_usage_supplied,include",
        [
            ({"tensorflow": ["tensorflow.keras.layers.Embedding"]}, None, False),
            (None, {"tensorflow": ["tensorflow.keras.layers.Embedding"]}, True),
            ({"tensorflow": ["tensorflow.keras.layers.Embedding"]}, {"flask": ["flask.Flask"]}, False),
            (
                {"tensorflow": ["tensorflow.keras.layers.Embedding", "tensorflow.Graph"]},
                {"tensorflow": ["tensorflow.Assert"]},
                False,
            ),
            ({"flask": ["flask.Flask"]}, {"tensorflow": ["tensorflow.keras.layers.Embedding"]}, False),
            ({"flask": ["flask.*"]}, {"flask": ["flask.Flask"]}, True),
            (
                {"flask": ["flask.*"]},
                {"tensorflow": ["tensorflow.keras.layers.Embedding"], "flask": ["flask.Flask"]},
                True,
            ),
            (
                {"flask": ["flask.*"], "tensorflow": ["tensorflow.*"]},
                {"flask": ["flask.Flask"], "tensorflow": ["tensorflow.keras.layers.Embedding"]},
                True,
            ),
        ],
    )
    def test_should_include_library_usage(
        self,
        builder_context: PipelineBuilderContext,
        library_usage_expected: Optional[Dict[str, Any]],
        library_usage_supplied: Optional[Dict[str, Any]],
        include: bool,
    ) -> None:
        """Test checking library usage."""
        should_include = {"adviser_pipeline": True}
        if library_usage_expected is not None:
            should_include["library_usage"] = library_usage_expected

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        UnitPrescription._PRESCRIPTION = {
            "name": "LibraryUsageUnit",
            "should_include": should_include,
        }

        builder_context.library_usage = library_usage_supplied
        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None
        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) == include

    @pytest.mark.parametrize(
        "labels_expected,labels_supplied,include",
        [
            ({}, {}, True),
            ({"foo": "bar"}, {}, False),
            ({}, {"foo": "bar"}, True),
            ({"foo": "bar"}, {"foo": "bar"}, True),
            ({"foo": "bar", "qux": "baz"}, {"foo": "bar"}, True),
            ({"foo": "bar"}, {"foo": "bar", "qux": "baz"}, True),
            ({"foo": "bar"}, {"qux": "baz", "baz": "foo"}, False),
        ],
    )
    def test_should_include_labels(
        self,
        builder_context: PipelineBuilderContext,
        labels_expected: Optional[Dict[str, str]],
        labels_supplied: Dict[str, str],
        include: bool,
    ) -> None:
        """Test including pipeline units based on labels."""
        should_include = {"adviser_pipeline": True}
        if labels_expected is not None:
            should_include["labels"] = labels_expected

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        UnitPrescription._PRESCRIPTION = {
            "name": "LabelsUnit",
            "should_include": should_include,
        }

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None
        builder_context.labels = labels_supplied
        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) == include

    @pytest.mark.parametrize(
        "abi_present,abi_configured,include",
        [
            ([], [], True),
            (["GLIBC_2.2.5"], [], True),
            ([], ["GLIBC_2.2.5"], False),
            (["GLIBC_2.2.5"], ["GLIBC_2.2.5"], True),
            (["GLIBC_2.2.5", "GNUTLS_3_6_14"], ["GLIBC_2.2.5"], True),
            (["GLIBC_2.2.5"], ["GNUTLS_3_6_14", "GLIBC_2.2.5"], False),
            (["GLIBC_2.2.5", "GNUTLS_3_6_14"], ["GNUTLS_3_6_14", "GLIBC_2.2.5"], True),
            (["GLIBC_2.2.5", "GNUTLS_3_6_14"], {"not": ["GNUTLS_3_6_14"]}, False),
            (["GLIBC_2.2.5"], {"not": ["GNUTLS_3_6_14"]}, True),
            (["GNUTLS_3_6_14"], {"not": ["GLIBC_2.2.5", "GNUTLS_3_6_14"]}, False),
            ([], {"not": ["GLIBC_2.2.5"]}, False),
        ],
    )
    def test_should_include_abi(
        self,
        builder_context: PipelineBuilderContext,
        abi_present: Optional[List[str]],
        abi_configured: Union[List[str], Dict[str, List[str]]],
        include: bool,
    ) -> None:
        """Test including pipeline units based on ABI present in the base image."""
        base_image_present_name, base_image_present_version = "s2i-thoth", "1.0.0"
        builder_context.project.runtime_environment.base_image = (
            f"{base_image_present_name}:v{base_image_present_version}"
        )
        should_include = {
            "adviser_pipeline": True,
            "runtime_environments": {"base_images": [builder_context.project.runtime_environment.base_image]},
        }

        if abi_configured:
            should_include["runtime_environments"]["abi"] = abi_configured

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        if abi_configured:
            builder_context.graph.should_receive("get_thoth_s2i_analyzed_image_symbols_all").with_args(
                base_image_present_name,
                base_image_present_version,
                is_external=False,
            ).and_return(abi_present).once()
        else:
            builder_context.graph.should_receive("get_thoth_s2i_analyzed_image_symbols_all").times(0)

        UnitPrescription._PRESCRIPTION = {
            "name": "SharedObjectsUnit",
            "should_include": should_include,
        }

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None

        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) == include

    @pytest.mark.parametrize("base_images", [[], ["s2i-thoth:v1.0.0"], ["s2i-thoth:v1.0.0", "s2i-thoth:v2.0.0"]])
    def test_should_include_abi_no_image_error(
        self, base_images: List[str], builder_context: PipelineBuilderContext
    ) -> None:
        """Test including pipeline units based on ABI present in the base image.

        This test covers a case when base image is not present in the user's configuration file.
        """
        builder_context.project.runtime_environment.base_image = None
        should_include = {
            "adviser_pipeline": True,
            "runtime_environments": {
                "abi": ["GNUTLS_3_6_14"],
            },
        }

        if base_images:
            should_include["runtime_environments"]["base_images"] = base_images

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        UnitPrescription._PRESCRIPTION = {
            "name": "SharedObjectsUnit",
            "should_include": should_include,
        }

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None

        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) is False

    @pytest.mark.parametrize(
        "rpms_present,rpms_configured,include",
        [
            ([], [], True),
            ([{"package_name": "git"}], [], True),
            ([], [{"package_name": "git"}], False),
            ([{"package_name": "git"}], [{"package_name": "git"}], True),
            ([{"package_name": "git"}], [{"package_name": "git"}], True),
            ([{"package_name": "git"}, {"package_name": "glibc"}], [{"package_name": "git"}], True),
            (
                _EXAMPLE_RPM_PACKAGES_PRESENT,
                [
                    {
                        "arch": "x86_64",
                        "package_name": "gcc",
                        "package_version": "8.3.1",
                        "release": "5.1.el8",
                    },
                    {
                        "epoch": "1",
                        "package_name": "findutils",
                        "package_version": "4.6.0",
                        "release": "20.el8",
                        "src": False,
                    },
                ],
                True,
            ),
            (
                _EXAMPLE_RPM_PACKAGES_PRESENT,
                [
                    {
                        "arch": "x86_64",
                        "package_name": "gcc",
                        "package_version": "8.3.1",
                        "release": "5.1.el8",
                    },
                    {
                        "epoch": "1",
                        "package_name": "findutils",
                        "package_version": "5.0.0",  # This version is not present.
                        "release": "20.el8",
                        "src": False,
                    },
                ],
                False,
            ),
            (
                _EXAMPLE_RPM_PACKAGES_PRESENT,
                {
                    "not": [
                        {
                            "arch": "x86_64",
                            "package_name": "gcc",
                            "package_version": "8.3.1",
                            "release": "5.1.el8",
                        },
                        {
                            "epoch": "1",
                            "package_name": "findutils",
                            "package_version": "5.0.0",  # This version is not present.
                            "release": "20.el8",
                            "src": False,
                        },
                    ]
                },
                True,
            ),
            (
                _EXAMPLE_RPM_PACKAGES_PRESENT,
                {
                    "not": [
                        {
                            "arch": "x86_64",
                            "package_name": "gcc",
                            "package_version": "8.3.1",
                            "release": "5.1.el8",
                        },
                        {
                            "epoch": "1",
                            "package_name": "findutils",
                            "package_version": "4.6.0",
                            "release": "20.el8",
                            "src": False,
                        },
                    ]
                },
                False,
            ),
        ],
    )
    def test_should_include_rpm_packages(
        self,
        builder_context: PipelineBuilderContext,
        rpms_present: List[Dict[str, str]],
        rpms_configured: List[Dict[str, str]],
        include: bool,
    ) -> None:
        """Test including pipeline units based on RPM packages present in the base image."""
        base_image_present_name, base_image_present_version = "s2i-thoth", "1.0.0"
        builder_context.project.runtime_environment.base_image = (
            f"{base_image_present_name}:v{base_image_present_version}"
        )
        should_include = {
            "adviser_pipeline": True,
            "runtime_environments": {"base_images": [builder_context.project.runtime_environment.base_image]},
        }

        if rpms_configured:
            should_include["runtime_environments"]["rpm_packages"] = rpms_configured

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        if rpms_configured:
            builder_context.graph.should_receive("get_last_analysis_document_id").with_args(
                base_image_present_name,
                base_image_present_version,
                is_external=False,
            ).and_return("package-extract-foo-bar").once()
            builder_context.graph.should_receive("get_rpm_package_version_all").with_args(
                "package-extract-foo-bar"
            ).and_return(rpms_present).once()
        else:
            builder_context.graph.should_receive("get_last_analysis_document_id").times(0)
            builder_context.graph.should_receive("get_rpm_package_version_all").times(0)

        UnitPrescription._PRESCRIPTION = {
            "name": "RPMPackagesUnit",
            "should_include": should_include,
        }

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None

        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) == include

    @pytest.mark.parametrize("base_images", [[], ["s2i-thoth:v1.0.0"], ["s2i-thoth:v1.0.0", "s2i-thoth:v2.0.0"]])
    def test_should_include_rpm_packages_no_image_error(
        self,
        builder_context: PipelineBuilderContext,
        base_images: List[str],
    ) -> None:
        """Test including pipeline units based on RPM packages present in the base image.

        This test covers a case when base image is not present in the user's configuration file.
        """
        builder_context.project.runtime_environment.base_image = None
        should_include = {
            "adviser_pipeline": True,
            "runtime_environments": {
                "rpm_packages": [
                    {
                        "package_name": "git",
                    }
                ],
            },
        }

        if base_images:
            should_include["runtime_environments"]["base_images"] = base_images

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        UnitPrescription._PRESCRIPTION = {
            "name": "RPMPackagesUnit",
            "should_include": should_include,
        }

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None

        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) is False

    def test_should_include_rpm_packages_no_analysis(self, builder_context: PipelineBuilderContext) -> None:
        """Test not including pipeline units if the base image was not analysed."""
        base_image_present_name, base_image_present_version = "s2i-thoth", "1.0.0"
        builder_context.project.runtime_environment.base_image = (
            f"{base_image_present_name}:v{base_image_present_version}"
        )
        should_include = {
            "adviser_pipeline": True,
            "runtime_environments": {
                "base_images": [builder_context.project.runtime_environment.base_image],
                "rpm_packages": [{"package_name": "glibc"}],
            },
        }

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        builder_context.graph.should_receive("get_last_analysis_document_id").with_args(
            base_image_present_name,
            base_image_present_version,
            is_external=False,
        ).and_return(None).once()
        builder_context.graph.should_receive("get_rpm_package_version_all").times(0)

        UnitPrescription._PRESCRIPTION = {
            "name": "RPMPackagesUnit",
            "should_include": should_include,
        }

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None

        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) is False

    @pytest.mark.parametrize(
        "python_present,python_configured,include",
        [
            ([], [], True),
            (_EXAMPLE_PYTHON_PACKAGES_PRESENT, [], True),
            # # Present.
            (_EXAMPLE_PYTHON_PACKAGES_PRESENT, [{"name": "dbus-python"}], True),
            (_EXAMPLE_PYTHON_PACKAGES_PRESENT, [{"name": "dbus-python"}, {"name": "wheel"}], True),
            (_EXAMPLE_PYTHON_PACKAGES_PRESENT, [{"name": "dbus-python", "version": "<1.3.0,>1.2.0"}], True),
            (
                _EXAMPLE_PYTHON_PACKAGES_PRESENT,
                [{"name": "dbus-python", "version": "<1.3.0,>1.2.0"}, {"name": "wheel", "version": "~=0.36.0"}],
                True,
            ),
            (
                _EXAMPLE_PYTHON_PACKAGES_PRESENT,
                [
                    {"name": "dbus-python", "version": "<1.3.0,>1.2.0", "location": "^/usr/lib64/.*"},
                    {"name": "wheel", "version": "<=0.40.0", "location": "^/opt/app-root/lib/python3.8/site-packages$"},
                ],
                True,
            ),
            (_EXAMPLE_PYTHON_PACKAGES_PRESENT, [{"name": "micropipenv"}], False),  # Not present.
            # Present but different version.
            (_EXAMPLE_PYTHON_PACKAGES_PRESENT, [{"name": "dbus-python", "version": ">=2.0.0"}], False),
            (
                _EXAMPLE_PYTHON_PACKAGES_PRESENT,
                [{"name": "dbus-python", "version": "<1.3.0,>1.2.0"}, {"name": "wheel", "version": "==0.36.0"}],
                False,
            ),
            # Present but in another location.
            (
                _EXAMPLE_PYTHON_PACKAGES_PRESENT,
                [
                    {"name": "dbus-python", "version": "<1.3.0,>1.2.0", "location": "^/usr/lib64/"},
                    {"name": "wheel", "version": "<=0.40.0", "location": "^/home/thoth$"},
                ],
                False,
            ),
            # Not present.
            ([], [{"name": "micropipenv"}], False),
            # Test "not" logic.
            # micropipenv is not present.
            (_EXAMPLE_PYTHON_PACKAGES_PRESENT, {"not": [{"name": "micropipenv"}]}, True),
            # All not present.
            (_EXAMPLE_PYTHON_PACKAGES_PRESENT, {"not": [{"name": "micropipenv"}, {"name": "selinon"}]}, True),
            # dbus-python present.
            (_EXAMPLE_PYTHON_PACKAGES_PRESENT, {"not": [{"name": "dbus-python"}, {"name": "selinon"}]}, False),
            # dbus-python present in different location, selinon not present
            (
                _EXAMPLE_PYTHON_PACKAGES_PRESENT,
                {
                    "not": [
                        {"name": "selinon"},
                        {"name": "dbus-python", "location": "^/home/thoth/.*"},
                    ]
                },
                True,
            ),
            # Version does not match so should be included based on "not".
            (
                _EXAMPLE_PYTHON_PACKAGES_PRESENT,
                {
                    "not": [
                        {"name": "dbus-python", "version": ">=2.0"},
                    ],
                },
                True,
            ),
        ],
    )
    def test_should_include_python_packages(
        self,
        builder_context: PipelineBuilderContext,
        python_present: List[Dict[str, str]],
        python_configured: List[Dict[str, str]],
        include: bool,
    ) -> None:
        """Test including pipeline units based on RPM packages present in the base image."""
        base_image_present_name, base_image_present_version = "s2i-thoth", "1.0.0"
        builder_context.project.runtime_environment.base_image = (
            f"{base_image_present_name}:v{base_image_present_version}"
        )
        should_include = {
            "adviser_pipeline": True,
            "runtime_environments": {"base_images": [builder_context.project.runtime_environment.base_image]},
        }

        if python_configured:
            should_include["runtime_environments"]["python_packages"] = python_configured

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        if python_configured:
            builder_context.graph.should_receive("get_last_analysis_document_id").with_args(
                base_image_present_name,
                base_image_present_version,
                is_external=False,
            ).and_return("package-extract-foo-bar").once()
            builder_context.graph.should_receive("get_python_package_version_all").with_args(
                "package-extract-foo-bar"
            ).and_return(python_present).once()
        else:
            builder_context.graph.should_receive("get_last_analysis_document_id").times(0)
            builder_context.graph.should_receive("get_rpm_package_version_all").times(0)

        UnitPrescription._PRESCRIPTION = {
            "name": "PythonPackagesUnit",
            "should_include": should_include,
        }

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None

        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) == include

    @pytest.mark.parametrize("base_images", [[], ["s2i-thoth:v1.0.0"], ["s2i-thoth:v1.0.0", "s2i-thoth:v2.0.0"]])
    def test_should_include_python_packages_no_image_error(
        self,
        builder_context: PipelineBuilderContext,
        base_images: List[str],
    ) -> None:
        """Test including pipeline units based on Python packages present in the base image.

        This test covers a case when base image is not present in the user's configuration file.
        """
        builder_context.project.runtime_environment.base_image = None
        should_include = {
            "adviser_pipeline": True,
            "runtime_environments": {
                "python_packages": [
                    {
                        "name": "dbus-python",
                    }
                ],
            },
        }

        if base_images:
            should_include["runtime_environments"]["base_images"] = base_images

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        UnitPrescription._PRESCRIPTION = {
            "name": "PythonPackagesUnit",
            "should_include": should_include,
        }

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None

        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) is False

    def test_should_include_python_packages_no_analysis(self, builder_context: PipelineBuilderContext) -> None:
        """Test not including pipeline units if the base image was not analyzed."""
        base_image_present_name, base_image_present_version = "s2i-thoth", "1.0.0"
        builder_context.project.runtime_environment.base_image = (
            f"{base_image_present_name}:v{base_image_present_version}"
        )
        should_include = {
            "adviser_pipeline": True,
            "runtime_environments": {
                "base_images": [builder_context.project.runtime_environment.base_image],
                "python_packages": [{"name": "dbus-python"}],
            },
        }

        PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA(should_include)

        builder_context.graph.should_receive("get_last_analysis_document_id").with_args(
            base_image_present_name,
            base_image_present_version,
            is_external=False,
        ).and_return(None).once()
        builder_context.graph.should_receive("get_python_package_version_all").times(0)

        UnitPrescription._PRESCRIPTION = {
            "name": "PythonPackagesUnit",
            "should_include": should_include,
        }

        builder_context.recommendation_type = RecommendationType.LATEST
        builder_context.decision_type = None

        assert builder_context.is_adviser_pipeline()
        builder_context.should_receive("is_included").with_args(UnitPrescription).and_return(False).once()
        assert UnitPrescription._should_include_base(builder_context) is False



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Recommend a specific TensorFlow based on API usage."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Set
from typing import TYPE_CHECKING
import attr
import orjson
import logging
import os

from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from voluptuous import Any as SchemaAny
from voluptuous import Required
from voluptuous import Schema

from ...enums import RecommendationType
from ...sieve import Sieve


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class TensorFlowAPISieve(Sieve):
    """A sieve that makes sure the right TensorFlow release is used based on user's API usage."""

    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): SchemaAny(str)})
    CONFIGURATION_DEFAULT = {"package_name": "tensorflow"}
    _LINK_API = jl("tf_api")
    _LINK_NO_API = jl("tf_no_api")

    _messages_logged = attr.ib(type=Set[str], factory=set, init=False)
    _no_api_logged = attr.ib(type=bool, default=False, init=False)
    _acceptable_releases = attr.ib(type=Optional[Set[str]], default=None, init=False)

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self._messages_logged.clear()
        self._acceptable_releases = None
        self._no_api_logged = False
        super().pre_run()

    def _pre_compute_releases(self) -> None:
        """Pre-compute releases that match library usage supplied by the user."""
        with open(os.path.join(self._DATA_DIR, "tensorflow", "api.json"), "r") as api_file:
            known_api = orjson.loads(api_file.read())

        self._acceptable_releases = set()
        tf_api_used = set(
            i for i in ((self.context.library_usage.get("report") or {}).get("tensorflow") or [])  # type: ignore
        )
        for tf_version, tf_api in known_api.items():
            if tf_api_used.issubset(tf_api):
                self._acceptable_releases.add(tf_version)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register this pipeline unit for adviser library usage is provided."""
        if (
            builder_context.is_adviser_pipeline()
            and not builder_context.is_included(cls)
            and builder_context.library_usage
            and builder_context.recommendation_type not in (RecommendationType.LATEST, RecommendationType.TESTING)
            and "tensorflow" in (builder_context.library_usage.get("report") or {})
        ):
            yield {"package_name": "tensorflow"}
            yield {"package_name": "tensorflow-gpu"}
            yield {"package_name": "intel-tensorflow"}
            yield {"package_name": "tensorflow-cpu"}
            return None

        yield from ()
        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Use specific TensorFlow release based on library usage as supplied by the user."""
        if self._acceptable_releases is None:
            self._pre_compute_releases()

        if not self._acceptable_releases:
            if not self._no_api_logged:
                self._no_api_logged = True
                msg = "No TensorFlow symbols API found in the database that would match TensorFlow symbols used"
                _LOGGER.warning("%s - see %s", msg, self._LINK_NO_API)
                self.context.stack_info.append(
                    {
                        "type": "WARNING",
                        "message": msg,
                        "link": self._LINK_NO_API,
                    }
                )

            yield from package_versions
            return

        for package_version in package_versions:
            version = ".".join(map(str, package_version.semantic_version.release[:2]))
            if version in self._acceptable_releases:
                yield package_version
            elif version not in self._messages_logged:
                self._messages_logged.add(version)
                msg = (
                    f"Removing TensorFlow {package_version.to_tuple()!r} as it does not provide required symbols "
                    f"in the exposed API"
                )
                _LOGGER.warning("%s - see %s", msg, self._LINK_API)
                self.context.stack_info.append({"type": "WARNING", "message": msg, "link": self._LINK_API})



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test a boot that checks for Pipfile hash and reports any mismatch to users.."""

from thoth.adviser.boots import PipfileHashBoot
from thoth.adviser.context import Context
from thoth.python import Project
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from ..base import AdviserUnitTestCase


class TestPipifleHashBoot(AdviserUnitTestCase):
    """Test a boot that checks for Pipfile hash and reports any mismatch to users.."""

    UNIT_TESTED = PipfileHashBoot

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        self.verify_multiple_should_include(builder_context)

    def test_no_include_no_pipenv_files(self, builder_context: PipelineBuilderContext) -> None:
        """Test if Pipenv files are not supplied - no hash is available to check against."""
        requirements_in = str(self.data_dir / "projects" / "requirements.in")
        requirements_txt = str(self.data_dir / "projects" / "requirements.txt")

        project = Project.from_pip_compile_files(
            requirements_path=requirements_in,
            requirements_lock_path=requirements_txt,
        )

        builder_context.project = project
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test registering this unit."""
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

    def test_run_hash_match(self, context: Context) -> None:
        """Test no stack info addition if hash mismatches."""
        assert not context.stack_info

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.run()

        assert len(context.stack_info) == 0
        assert context.stack_info == []

    def test_run_hash_mismatch(self, context: Context) -> None:
        """Test adding stack info if hash matches."""
        assert not context.stack_info
        context.project.pipfile_lock.meta.hash["sha256"] = "foo"

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.run()

        assert len(context.stack_info) == 2
        assert set(context.stack_info[0].keys()) == {"link", "message", "type"}
        assert context.stack_info[0]["message"] == (
            "Pipfile hash stated in the Pipfile.lock (foo) does not correspond to the "
            "hash computed (001e85) - was Pipfile adjusted?"
        )

        assert set(context.stack_info[1].keys()) == {"link", "message", "type"}
        assert (
            context.stack_info[1]["message"]
            == "Detected changes in the lock file invalidate using user's stack as a base"
        )
        assert self.verify_justification_schema(context.stack_info)

    def test_run_dev(self, context: Context) -> None:
        """Test discarding user's lock file if development dependencies are not present in lock but provided."""
        pipfile = str(self.data_dir / "projects" / "dev" / "Pipfile")
        pipfile_lock = str(self.data_dir / "projects" / "dev" / "Pipfile.lock")

        project = Project.from_files(pipfile_path=pipfile, pipfile_lock_path=pipfile_lock)

        assert project.pipfile.dev_packages.packages
        assert project.pipfile_lock.dev_packages.packages

        # Remove packages.
        project.pipfile_lock.dev_packages.packages.clear()
        context.cli_parameters["dev"] = True

        context.project = project

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.run()

        assert project.pipfile.dev_packages.packages
        assert not project.pipfile_lock, "Lock file was not removed from the input"
        assert len(context.stack_info) == 1
        assert self.verify_justification_schema(context.stack_info)

    def test_run_dev_no_dev(self, context: Context) -> None:
        """Test discarding user's lock file if development dependencies are present but no --dev flag provided."""
        pipfile = str(self.data_dir / "projects" / "dev" / "Pipfile")
        pipfile_lock = str(self.data_dir / "projects" / "dev" / "Pipfile.lock")

        project = Project.from_files(pipfile_path=pipfile, pipfile_lock_path=pipfile_lock)

        assert project.pipfile.dev_packages.packages
        assert project.pipfile_lock.dev_packages.packages

        context.cli_parameters["dev"] = False
        context.project = project
        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.run()

        assert project.pipfile.dev_packages.packages
        assert not project.pipfile_lock, "Lock file was not removed from the input"
        assert len(context.stack_info) == 1
        assert self.verify_justification_schema(context.stack_info)

    def test_run_dev_noop(self, context: Context) -> None:
        """Test not discarding user's lock file if development dependencies are present in lock."""
        pipfile = str(self.data_dir / "projects" / "dev" / "Pipfile")
        pipfile_lock = str(self.data_dir / "projects" / "dev" / "Pipfile.lock")

        project = Project.from_files(pipfile_path=pipfile, pipfile_lock_path=pipfile_lock)

        assert project.pipfile.dev_packages.packages
        assert project.pipfile_lock.dev_packages.packages

        context.project = project
        context.cli_parameters["dev"] = True
        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.run()

        assert project.pipfile.dev_packages.packages
        assert project.pipfile_lock
        assert project.pipfile_lock.dev_packages.packages
        assert len(context.stack_info) == 0

    def test_run_dev_noop_no_dev(self, context: Context) -> None:
        """Test not discarding user's lock file if development dependencies are present in lock."""
        pipfile = str(self.data_dir / "projects" / "dev" / "Pipfile")
        pipfile_lock = str(self.data_dir / "projects" / "dev" / "Pipfile.lock")

        project = Project.from_files(pipfile_path=pipfile, pipfile_lock_path=pipfile_lock)

        assert project.pipfile.dev_packages.packages
        assert project.pipfile_lock.dev_packages.packages

        context.project = project
        project.pipfile_lock.dev_packages.packages.clear()
        context.cli_parameters["dev"] = False
        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.run()

        assert project.pipfile.dev_packages.packages
        assert project.pipfile_lock
        assert len(context.stack_info) == 0



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Representation of an advised stack."""

import os
import json
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import attr

from thoth.common import RuntimeEnvironment
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from thoth.python import Project
from thoth.python import PipfileLock
from thoth.storages.exceptions import NotFoundError

from .context import Context
from .state import State
from .utils import log_once


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True, eq=False, order=False)
class Product:
    """A representation of an advised stack."""

    _LOG_HASHES: Set[Tuple[str, str, str]] = set()
    _NO_OBSERVATION_JUSTIFICATION: Dict[str, str] = {
        "link": jl("no_observations"),
        "message": "No issues spotted for this stack based on Thoth's database",
        "type": "INFO",
    }

    # This project is "advised project" and does not need to be exactly the same as the one stated in context.
    project = attr.ib(type=Project, kw_only=True)
    score = attr.ib(type=float, kw_only=True)
    justification = attr.ib(type=List[Dict[str, str]], kw_only=True)
    advised_runtime_environment = attr.ib(type=Optional[RuntimeEnvironment], default=None, kw_only=True)
    advised_manifest_changes = attr.ib(type=List[List[Dict[str, Any]]], kw_only=True, default=attr.Factory(list))
    _context = attr.ib(type=Context, kw_only=True)

    @classmethod
    def from_final_state(cls, *, context: Context, state: State) -> "Product":
        """Instantiate advised stack from final state produced by adviser's pipeline."""
        assert state.is_final(), "Instantiating product from a non-final state"

        package_versions_locked = []
        for package_tuple in state.resolved_dependencies.values():
            package_version: PackageVersion = context.get_package_version(package_tuple, graceful=False)

            # Fill package hashes before instantiating the final product.
            if not package_version.hashes:
                # We can re-use already existing package-version - in that case it already keeps hashes from
                # a previous product instantiation.
                hashes = context.graph.get_python_package_hashes_sha256(*package_tuple)
                package_version.hashes = ["sha256:" + h for h in hashes]

                if not package_version.hashes:
                    log_once(_LOGGER, cls._LOG_HASHES, package_tuple, "No hashes found for package %r", package_tuple)

            direct_dependency = context.project.pipfile.packages.packages.get(package_version.name)
            if direct_dependency:
                # If the given package is a direct dependency, reuse markers stated by the user.
                package_version.markers = direct_dependency.markers
            else:
                # Fill environment markers by checking dependencies that introduced this dependency.
                # We do it only if we have no hashes - if hashes are present, the environment marker was
                # already picked (can be set to None if no marker is present).
                # For direct dependencies, dependents can return an empty set (if dependency is not
                # shared with other dependencies) and marker is propagated from PackageVersion registered in
                # Context.register_package_version.
                dependents_tuples = context.dependents[package_tuple[0]][package_tuple]

                # Marker depends based on the stack that was resolved. Do not change package_version directly,
                # rather clone it and used a cloned version not to clash with environment markers.
                environment_markers = []
                for dependent_tuple in dependents_tuples:
                    try:
                        marker = context.graph.get_python_environment_marker(
                            *dependent_tuple[0],
                            dependency_name=package_tuple[0],
                            dependency_version=package_tuple[1],
                            os_name=dependent_tuple[1],
                            os_version=dependent_tuple[2],
                            python_version=dependent_tuple[3],
                        )
                    except NotFoundError:
                        # This can happen if we do resolution that is agnostic to runtime
                        # environment. In that case a dependency introduced in one runtime
                        # environment does not need to co-exist in another runtime environment considering
                        # marker evaluation result.
                        continue

                    if marker and marker not in environment_markers:
                        environment_markers.append(marker)
                    elif not marker:
                        # One or multiple dependencies require this dependency to be always present, clear any
                        # environment markers.
                        environment_markers.clear()
                        break

                if environment_markers:
                    if len(environment_markers) > 1:
                        markers = " or ".join(f"({m})" for m in environment_markers)
                    else:
                        markers = environment_markers[0]

                    package_version = attr.evolve(
                        package_version,
                        markers=markers,
                    )

            package_versions_locked.append(package_version)

        advised_project = Project.from_package_versions(
            packages=list(context.project.iter_dependencies(with_devel=True)),
            packages_locked=package_versions_locked,
            meta=context.project.pipfile.meta,
            runtime_environment=context.project.runtime_environment,
        )

        # Keep thoth section untouched.
        advised_project.pipfile.thoth = context.project.pipfile.thoth

        justification_metadata: List[Dict[str, Any]] = []
        metadata = os.getenv("THOTH_ADVISER_METADATA")
        if metadata:
            try:
                metadata_content = json.loads(metadata)
                justification_metadata = (metadata_content.get("thoth.adviser") or {}).get("justification") or []
            except Exception:
                _LOGGER.exception("Failed to parse adviser metadata")

        justification = justification_metadata + state.justification
        if not justification:
            justification.append(cls._NO_OBSERVATION_JUSTIFICATION)

        return cls(
            project=advised_project,
            score=state.score,
            justification=justification,
            advised_runtime_environment=state.advised_runtime_environment,
            advised_manifest_changes=state.advised_manifest_changes,
            context=context,
        )

    @staticmethod
    def _construct_dependency_graph(context: Context, pipfile_lock: PipfileLock) -> Dict[str, Any]:
        """Construct dependency graph for the given Pipfile.lock."""
        nodes: List[str] = []
        nodes_idx: Dict[Tuple[str, str, str], int] = {}
        for package_version in pipfile_lock.packages.packages.values():
            package_version_tuple = package_version.to_tuple()
            nodes.append(package_version_tuple)
            nodes_idx[package_version_tuple] = len(nodes_idx)

        edges = []
        for src_idx, package_version_tuple in enumerate(nodes):
            dependencies = context.dependencies.get(package_version_tuple[0], {}).get(package_version_tuple)
            if not dependencies:
                continue

            for dependency in dependencies:
                dst_idx = nodes_idx.get(dependency)
                if dst_idx is not None:
                    edges.append([src_idx, dst_idx])

        # Discard version and index url as package name is good enough to locate nodes.
        return {"nodes": [i[0] for i in nodes], "edges": edges}

    def to_dict(self) -> Dict[str, Any]:
        """Convert this instance into a dictionary."""
        advised_runtime_environment = None
        if self.advised_runtime_environment:
            advised_runtime_environment = self.advised_runtime_environment.to_dict()

        return {
            "project": self.project.to_dict(keep_thoth_section=True),
            "dependency_graph": self._construct_dependency_graph(self._context, self.project.pipfile_lock),
            "score": self.score,
            "justification": self.justification,
            "advised_runtime_environment": advised_runtime_environment,
            "advised_manifest_changes": self.advised_manifest_changes,
        }



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of Adaptive Simulated Annealing (ASA)."""

from typing import Callable

import flexmock

from hypothesis import given
from hypothesis.strategies import integers
from hypothesis.strategies import floats

from thoth.adviser.beam import Beam
from thoth.adviser.predictors import AdaptiveSimulatedAnnealing
from thoth.adviser.state import State

from ..base import AdviserTestCase


class TestAdaptiveSimulatedAnnealing(AdviserTestCase):
    """Tests related to Adaptive Simulated Annealing (ASA)."""

    @given(
        floats(min_value=0.0, allow_nan=False, allow_infinity=False),
        integers(min_value=0),
        integers(min_value=1),
        integers(min_value=0),
        integers(min_value=0),
    )
    def test_temperature_function(
        self,
        t0: float,
        accepted_final_states_count: int,
        limit: int,
        iteration: int,
        count: int,
    ) -> None:
        """Test the temperature function never drops bellow 0."""
        context = flexmock(
            accepted_final_states_count=accepted_final_states_count,
            limit=limit,
            iteration=iteration,
            count=count,
            beam=flexmock(size=96),
        )

        predictor = AdaptiveSimulatedAnnealing()
        assert predictor._temperature_function(t0=t0, context=context) >= 0.0, "Temperature dropped bellow 0 or is NaN"

    @given(
        floats(allow_nan=False, allow_infinity=False),
        floats(allow_nan=False, allow_infinity=False),
        floats(min_value=0.0, allow_nan=False, allow_infinity=False),
    )
    def test_acceptance_probability(self, top_score: float, neighbour_score: float, temperature: float) -> None:
        """Test acceptance probability is always between 0 and 1."""
        acceptance_probability = AdaptiveSimulatedAnnealing._compute_acceptance_probability(
            top_score=top_score,
            neighbour_score=neighbour_score,
            temperature=temperature,
        )
        assert 0.0 <= acceptance_probability <= 1.0, "Acceptance probability not within 0 and 1"

    def test_pre_run(self) -> None:
        """Test pre-run initialization."""
        context = flexmock(limit=100)

        predictor = AdaptiveSimulatedAnnealing()
        assert predictor._temperature == 0.0
        predictor._temperature_history = [(0.1, False, 0.2, 3), (0.42, True, 0.66, 47)]

        with predictor.assigned_context(context):
            predictor.pre_run()

            assert predictor._temperature == context.limit, "Predictor's limit not initialized correctly"
            assert predictor._temperature_history == [], "Predictor's temperature history no discarded"

    @given(
        integers(min_value=1, max_value=256),
        integers(min_value=1, max_value=1024),
        integers(min_value=1, max_value=1024),
        integers(min_value=0, max_value=1024),
        integers(min_value=0, max_value=1024),
    )
    def test_run(
        self,
        state_factory: Callable[[], State],
        state_count: int,
        limit: int,
        count: int,
        iteration: int,
        accepted_final_states: int,
    ) -> None:
        """Test running the annealing."""
        beam = Beam()
        state = state_factory()
        for _ in range(state_count):
            cloned_state = state.clone()
            cloned_state.iteration = state.iteration + 1
            beam.add_state(cloned_state)

        predictor = AdaptiveSimulatedAnnealing()
        context = flexmock(
            accepted_final_states_count=accepted_final_states,
            count=count,
            iteration=iteration,
            limit=limit,
            beam=beam,
        )
        with predictor.assigned_context(context):
            next_state, package_tuple = predictor.run()
            assert next_state in beam.iter_states()
            assert package_tuple is not None
            assert package_tuple[0] in next_state.unresolved_dependencies
            assert package_tuple in next_state.unresolved_dependencies[package_tuple[0]].values()



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test solver implementation on top Thoth's knowledge graph."""

import pytest

from thoth.adviser.solver import PythonGraphSolver
from thoth.adviser.solver import PythonPackageGraphSolver
from thoth.adviser.solver import GraphReleasesFetcher
from thoth.adviser.solver import PackageVersionDependencyParser
from thoth.python import PackageVersion
from thoth.solver.python.base import SolverException

from .base import AdviserTestCase
from .graph_mock import MockedGraphDatabase


class TestSolver(AdviserTestCase):
    """Test Thoth's solver implementation on top of Thoth's knowledge graph."""

    @pytest.mark.parametrize("graph", [MockedGraphDatabase("db_0.yaml")])
    def test_db_0_raises(self, graph: MockedGraphDatabase) -> None:
        """Check that there is raised an exception if no releases were found."""
        with pytest.raises(SolverException):
            PythonGraphSolver(
                dependency_parser=PackageVersionDependencyParser(),
                releases_fetcher=GraphReleasesFetcher(graph=graph),
            ).solve(
                [
                    PackageVersion(
                        name="nonexisting-foo",
                        version="==1.0.0",
                        index=None,
                        develop=False,
                    )
                ],
                graceful=False,
            )

    @pytest.mark.parametrize("graph", [MockedGraphDatabase("db_0.yaml")])
    def test_db_0(self, graph: MockedGraphDatabase) -> None:
        """Check that resolving can gather all versions available in the graph database."""
        resolved = PythonGraphSolver(
            dependency_parser=PackageVersionDependencyParser(),
            releases_fetcher=GraphReleasesFetcher(graph=graph),
        ).solve(
            [PackageVersion(name="a", version="*", index=None, develop=False)],
            graceful=False,
        )
        assert len(resolved) == 1
        assert "a" in resolved
        assert set(resolved["a"]) == {
            ("1.0.0", "index1"),
            ("1.1.0", "index1"),
            ("1.2.0", "index2"),
        }

    @pytest.mark.parametrize("graph", [MockedGraphDatabase("db_0.yaml")])
    def test_db_0_multiple(self, graph: MockedGraphDatabase) -> None:
        """Check that resolving can resolve multiple Python packages."""
        resolved = PythonGraphSolver(
            dependency_parser=PackageVersionDependencyParser(),
            releases_fetcher=GraphReleasesFetcher(graph=graph),
        ).solve(
            [
                PackageVersion(name="a", version="*", index=None, develop=False),
                PackageVersion(name="b", version=">1.0.0", index=None, develop=False),
            ],
            graceful=False,
        )

        assert len(resolved) == 2

        assert "a" in resolved
        assert set(resolved["a"]) == {
            ("1.0.0", "index1"),
            ("1.1.0", "index1"),
            ("1.2.0", "index2"),
        }

        assert "b" in resolved
        assert set(resolved["b"]) == {("2.0.0", "index1"), ("3.0.0", "index2")}

    @pytest.mark.parametrize("graph", [MockedGraphDatabase("db_0.yaml")])
    def test_db_0_package_raises(self, graph: MockedGraphDatabase) -> None:
        """Check that there is raised an exception if no releases were found."""
        with pytest.raises(SolverException):
            PythonPackageGraphSolver(graph=graph).solve(
                [
                    PackageVersion(
                        name="nonexisting-foo",
                        version="==1.0.0",
                        index=None,
                        develop=False,
                    )
                ],
                graceful=False,
            )

    @pytest.mark.parametrize("graph", [MockedGraphDatabase("db_0.yaml")])
    def test_db_0_all_package_versions(self, graph: MockedGraphDatabase) -> None:
        """Check that resolving can gather all versions available in the graph database."""
        resolved = PythonPackageGraphSolver(graph=graph).solve(
            [PackageVersion(name="a", version="*", index=None, develop=False)],
            graceful=False,
        )

        assert len(resolved) == 1
        assert "a" in resolved
        # 1.0.0, 1.1.0, 1.2.0
        assert len(resolved["a"]) == 3

        assert all(package_version.name == "a" for package_version in resolved["a"])
        assert set(package_version.version for package_version in resolved["a"]) == {
            "==1.0.0",
            "==1.1.0",
            "==1.2.0",
        }

    @pytest.mark.parametrize("graph", [MockedGraphDatabase("db_0.yaml")])
    def test_db_0_package_multiple(self, graph: MockedGraphDatabase) -> None:
        """Check that resolving can resolve multiple Python packages."""
        resolved = PythonPackageGraphSolver(graph=graph).solve(
            [
                PackageVersion(name="a", version="*", index=None, develop=False),
                PackageVersion(name="b", version=">1.0.0", index=None, develop=False),
            ],
            graceful=False,
        )
        assert len(resolved) == 2
        assert "a" in resolved
        assert "b" in resolved

        assert all(package_version.name == "a" for package_version in resolved["a"])
        assert all(package_version.name == "b" for package_version in resolved["b"])
        assert set(package_version.version for package_version in resolved["a"]) == {
            "==1.0.0",
            "==1.1.0",
            "==1.2.0",
        }
        assert set(package_version.version for package_version in resolved["b"]) == {
            "==2.0.0",
            "==3.0.0",
        }

    @pytest.mark.parametrize("graph", [MockedGraphDatabase("db_0.yaml")])
    def test_db_0_multiple_times_error(self, graph: MockedGraphDatabase) -> None:
        """Check that resolving can resolve multiple Python packages."""
        from thoth.solver.python.base import SolverException

        with pytest.raises(SolverException):
            PythonPackageGraphSolver(graph=graph).solve(
                [
                    PackageVersion(name="a", version="<=1.2.0", index=None, develop=False),
                    PackageVersion(name="a", version=">1.0.0", index=None, develop=False),
                ],
                graceful=False,
            )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test using the right TensorFlow release based on library usage (symbols used) supplied."""

import pytest

from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import TensorFlowAPISieve
from thoth.adviser.context import Context
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserUnitTestCase


class TestTensorFlowAPISieve(AdviserUnitTestCase):
    """Test using the right TensorFlow release for specific CUDA releases."""

    UNIT_TESTED = TensorFlowAPISieve

    _PACKAGES_AFFECTED = [
        "tensorflow",
        "tensorflow-gpu",
        "intel-tensorflow",
        "tensorflow-cpu",
    ]

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.STABLE
        builder_context.library_usage = {"report": {"flask": ["flask.Flask"], "tensorflow": ["tensorflow."]}}

        pipeline_config = list(self.UNIT_TESTED.should_include(builder_context))
        assert len(self._PACKAGES_AFFECTED) == len(pipeline_config)

        for package_name in self._PACKAGES_AFFECTED:
            assert package_name
            assert {"package_name": package_name} in pipeline_config

        for item in pipeline_config:
            assert pipeline_config
            unit = self.UNIT_TESTED()
            unit.update_configuration(item)
            builder_context.add_unit(unit)

        assert list(self.UNIT_TESTED.should_include(builder_context)) == [], "The unit must not be included"

    def test_recommendation_types_considered(self) -> None:
        """Test recommendation types that were considered during implementation of this pipeline unit.

        This test will fail when a new recommendation type is added. In such case, review the pipeline unit
        implementation and adjust it accordingly.
        """
        assert set(RecommendationType) == {
            RecommendationType.STABLE,
            RecommendationType.TESTING,
            RecommendationType.LATEST,
            RecommendationType.PERFORMANCE,
            RecommendationType.SECURITY,
        }

    @pytest.mark.parametrize(
        "recommendation_type",
        [RecommendationType.STABLE, RecommendationType.PERFORMANCE, RecommendationType.SECURITY],
    )
    def test_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType) -> None:
        """Test including this pipeline unit."""
        builder_context.decision_type = None
        builder_context.recommendation_type = recommendation_type
        builder_context.library_usage = {"report": {"tensorflow": ["tensorflow.v2.__version__"]}}
        assert builder_context.is_adviser_pipeline()
        assert list(self.UNIT_TESTED.should_include(builder_context)) != []

    @pytest.mark.parametrize(
        "recommendation_type,decision_type,library_usage",
        [
            (RecommendationType.LATEST, None, {"report": {"tensorflow": ["tensorflow.__version__"]}}),
            (RecommendationType.TESTING, None, {"report": {"tensorflow": ["tensorflow.__version__"]}}),
            (None, DecisionType.RANDOM, {"report": {"tensorflow": ["tensorflow.__version__"]}}),  # Dependency Monkey
            (RecommendationType.STABLE, None, {"report": {"flask": "flask.Flask"}}),  # No tensorflow used.
            (RecommendationType.STABLE, None, None),  # No library usage.
        ],
    )
    def test_no_include(
        self,
        builder_context: PipelineBuilderContext,
        recommendation_type: RecommendationType,
        decision_type: DecisionType,
        library_usage: str,
    ) -> None:
        """Test not including this pipeline unit."""
        builder_context.decision_type = decision_type
        builder_context.recommendation_type = recommendation_type
        builder_context.library_usage = library_usage
        assert builder_context.is_adviser_pipeline() or builder_context.is_dependency_monkey_pipeline()
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_sieve(self, context: Context) -> None:
        """Test sieving TensorFlow based on API symbols."""
        context.library_usage = {
            # run_functions_eagerly available since 2.3.0.
            "report": {"tensorflow": ["tensorflow.v2.__version__", "tensorflow.raw_ops.LoadDataset"]}
        }
        source = Source("https://pypi.org/simple")
        pv_0 = PackageVersion(
            name="tensorflow-gpu",
            version="==2.2.0",
            develop=False,
            index=source,
        )
        pv_1 = PackageVersion(
            name="tensorflow",
            version="==2.2.0",
            develop=False,
            index=source,
        )
        pv_2 = PackageVersion(
            name="tensorflow",
            version="==2.3.0",
            develop=False,
            index=source,
        )
        pv_3 = PackageVersion(
            name="intel-tensorflow",
            version="==2.4.0",
            develop=False,
            index=source,
        )
        pv_4 = PackageVersion(
            name="tensorflow",
            version="==1.13.0",
            develop=False,
            index=source,
        )

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.pre_run()
            result = list(unit.run((pv for pv in (pv_0, pv_1, pv_2, pv_3, pv_4))))
            assert len(result) == 2
            assert result == [pv_2, pv_3]

    def test_unknown_symbol(self, context: Context) -> None:
        """Test no sieving is done if an unknown symbol is spotted."""
        context.library_usage = {
            # run_functions_eagerly available since 2.3.0.
            "report": {"tensorflow": ["tensorflow.SomeUnknownSymbol"]}
        }
        source = Source("https://pypi.org/simple")
        pv_0 = PackageVersion(
            name="tensorflow",
            version="==2.2.0",
            develop=False,
            index=source,
        )
        pv_1 = PackageVersion(
            name="tensorflow",
            version="==2.3.0",
            develop=False,
            index=source,
        )
        pv_2 = PackageVersion(
            name="tensorflow",
            version="==2.4.0",
            develop=False,
            index=source,
        )

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.pre_run()
            result = list(unit.run((pv for pv in (pv_0, pv_1, pv_2))))
            assert len(result) == 3
            assert result == [pv_0, pv_1, pv_2]

    def test_multiple_calls(self, context: Context) -> None:
        """Test multiple calls of resolver invalidate internal caches."""
        context.library_usage = {
            # run_functions_eagerly available since 2.3.0.
            "report": {"tensorflow": ["tensorflow.v2.__version__", "tensorflow.raw_ops.LoadDataset"]}
        }
        source = Source("https://pypi.org/simple")
        pv_1 = PackageVersion(
            name="tensorflow",
            version="==2.2.0",
            develop=False,
            index=source,
        )
        pv_2 = PackageVersion(
            name="tensorflow",
            version="==2.3.0",
            develop=False,
            index=source,
        )

        unit = self.UNIT_TESTED()
        with unit.assigned_context(context):
            unit.pre_run()
            result = list(unit.run((pv for pv in (pv_1, pv_2))))
            assert len(result) == 1
            assert result == [pv_2]

        # Now without symbols introduced in 2.3.0 on the same unit instance.
        context.library_usage = {
            # run_functions_eagerly available since 2.3.0.
            "report": {"tensorflow": ["tensorflow.v2.__version__"]}
        }

        with unit.assigned_context(context):
            unit.pre_run()
            result = list(unit.run((pv for pv in (pv_1, pv_2))))
            assert len(result) == 2
            assert result == [pv_1, pv_2]
